/* $Workfile: AlvisBoot.c $       $Revision: 6 $      $Modtime: 3/04/10 2:46p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    AlvisBoot.c

    An example console application demonstrating how to boot the DMPs on an
    Alvis-ASM board.
    
    Press 'h' or '?' for help. 'q' to quit.

*******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#if !defined(_WIN32_WCE)
#include <assert.h>
#include <time.h>
#endif  // #if !defined(_WIN32_WCE)

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxGeneric.h"
#include "OtxErr.h"
#include "OtxBrd.h"
#include "OtxDef.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxHdlc.h"
#include "OtxT1E1.h"
#include "OtxData.h"
#include "OtxDsp.h"

#define DEMO_DEFAULT_LOG_FILE "ThorAccess.log"
#define DEMO_DEFAULT_BOARD_NO 0


typedef struct DemoEnvT {
  char                  *pszInputStr;
  OTX_BOOL               bExitOnError;
  char                   cwd[OTX_MAX_PATH];
  char                   logFileName[OTX_MAX_PATH];
  OTX_INT32              nBoardType;
  int                    boardNo;
  int                    coDmp;
  int                    nAsmOffset;
} DemoEnvT;


/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoOpenPhysicalDevices(void);
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);
static void DemoHelpPanel(void);
static void DemoLogf(char *formatStr, ...);
static void DemoCheckRc(IN OTX_RESULT otxRc);
static void DemoExitQuestion(void);

static OTX_INT32 DemoProcessUserCommand(char commandChar);


/*******************************************************************************
 Global Data
*******************************************************************************/
OTX_EVENT   g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hEventQueue  = 0;
FILE       *g_pDemoLogFile = NULL;
DemoEnvT    g_demoEnv;

struct PhysDevs {
  OTX_HANDLE hThorController0;
  OTX_HANDLE hPlx;
} g_hPhysDevs;



/*******************************************************************************

  @func
    Open the Physical Devices on the OTX board

  @comm
    Calls function <f OtxDrvCreatePhysicalDevice> returning a handle to a
    Device Object representing a physical device, which can be the
    entire board or an integrated circuit (chip) on the board. The number of
    physical devices is set per board and cannot change dynamically.

*******************************************************************************/
void DemoOpenPhysicalDevices(void)
{
  OTX_RESULT result;

  // Connect to the Physical Board Controller
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_2_PCI_PLUS:
    DemoLogf("Opening Thor-2-PCI-Plus board\n");
    break;
  case OTX_DEVICE_THOR_2_PCI_EXPRESS:
    DemoLogf("Opening Thor-2-PCI-Express board\n");
    break;
  case OTX_DEVICE_THOR_1_PCI_PLUS:
    DemoLogf("Opening Thor-1-PCI-Plus board\n");
    break;
  case OTX_DEVICE_THOR_8_PCI_PLUS:
    DemoLogf("Opening Thor-8-PCI-Plus board\n");
    break;
  case OTX_DEVICE_THOR_4_PCI_PLUS:
    DemoLogf("Opening Thor-4-PCI-Plus board\n");
    break;
  case OTX_DEVICE_GIMLE_16_PCIE:
    DemoLogf("Opening Gimle-16-PCIe board\n");
    break;
  case OTX_DEVICE_GIMLE_16_PCI_PLUS:
    DemoLogf("Opening Gimle-16-PCI-Plus board\n");
    break;
  case OTX_DEVICE_GIMLE_16_PCI_BASIC_32:
    DemoLogf("Opening Gimle-16-PCI-Basic-32 board\n");
    break;
  case OTX_DEVICE_GIMLE_16_PCI_BASIC_64:
    DemoLogf("Opening Gimle-16-PCI-Basic-64 board\n");
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }

  result = OtxDrvGetDeviceHandle(0, (OtxPhysicalDeviceTypeE) g_demoEnv.nBoardType, g_demoEnv.boardNo, &g_hPhysDevs.hThorController0); DemoCheckRc(result);

  DemoLogf("Board Device Handle = 0x%.8X\n", g_hPhysDevs.hThorController0);

  result = OtxDrvGetDeviceHandle(g_hPhysDevs.hThorController0, (OtxPhysicalDeviceTypeE) OTX_DEVICE_PCI_BRIDGE, 0, &g_hPhysDevs.hPlx); 

  DemoLogf("Plx Device Handle = 0x%.8X\n", g_hPhysDevs.hPlx);

}



/*******************************************************************************

  @func

  Parse the program's command line arguments and calls upon
  requested actions.

  There are some optional switches that can be used following
  the driver name on the command line. Switches are are
  separated by spaces, and can appear in arbitrary order.
  Switches consist of a hyphen and a letter:

  @comm

*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP) 
{
  short  i;

  // Set reasonable defaults before reading from command line
  envP->pszInputStr       = NULL;
  envP->bExitOnError      = OTX_FALSE;
  envP->nBoardType        = OTX_DEVICE_THOR_8_PCI_PLUS;
  envP->boardNo           = DEMO_DEFAULT_BOARD_NO;
  envP->coDmp             = 4;
  envP->nAsmOffset        = 0x100;

  strcpy(envP->logFileName, envP->cwd);
  strcat(envP->logFileName, DEMO_DEFAULT_LOG_FILE);
  for (i = 1; i < argc; i++) {        // Scan command line arguments
    if (*(argv[i]+1) == 'Z') {   // Command input string -- case sensitive
      envP->pszInputStr = argv[i]+2;
    } else {    
      OsaStdStrUpr(argv[i]);
      if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
        if (*(argv[i]+1) == 'L') {   // Log file
          strcpy(envP->logFileName, argv[i] + 2);
        } else if (*(argv[i]+1) == 'Q') {   // Exit on error
          envP->bExitOnError = OTX_TRUE;
        } else if (*(argv[i]+1) == 'B') {   // Board number
          envP->boardNo = strtoul(argv[i]+2, NULL, 0);
        } else if (*(argv[i]+1) == 'T') {   // Board Type
          if (*(argv[i] + 2) == 'T') {
            envP->nBoardType  = OTX_DEVICE_THOR_2_PCI_PLUS;
            envP->nAsmOffset        = 0x80;
          } else if (*(argv[i] + 2) == 'E') {
            envP->nBoardType  = OTX_DEVICE_THOR_2_PCI_EXPRESS;
            envP->nAsmOffset        = 0x80;
          } else if (*(argv[i] + 2) == 'P') {
            envP->nBoardType  = OTX_DEVICE_THOR_8_PCI_PLUS;
            envP->nAsmOffset        = 0x100;  // For 2.0, but it should be 0x80 for the 1.0
          } else if (*(argv[i] + 2) == '4') {
            envP->nBoardType  = OTX_DEVICE_THOR_4_PCI_PLUS;
            envP->nAsmOffset        = 0x80;
          } else if (*(argv[i] + 2) == '1') {
            envP->nBoardType  = OTX_DEVICE_THOR_1_PCI_PLUS;
            envP->nAsmOffset        = 0x80;
          } else if (*(argv[i] + 2) == '9') {
            envP->nBoardType  = OTX_DEVICE_GIMLE_16_PCIE;
            envP->nAsmOffset        = 0x80;
          } else if (*(argv[i] + 2) == 'G') {
            envP->nBoardType  = OTX_DEVICE_GIMLE_16_PCI_PLUS;
            envP->nAsmOffset        = 0x80;
          } else if (*(argv[i] + 2) == '3') {
            envP->nBoardType  = OTX_DEVICE_GIMLE_16_PCI_BASIC_32;
            envP->nAsmOffset        = 0x80;
          } else if (*(argv[i] + 2) == '6') {
            envP->nBoardType  = OTX_DEVICE_GIMLE_16_PCI_BASIC_64;
            envP->nAsmOffset        = 0x80;
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            exit(1);
          }
        } else {
          DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          exit(1);
        }
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
			  exit(1);
      }
    }
  }
}


/*******************************************************************************

  @func

  Displays a help for the command line.

  @comm

*******************************************************************************/
static void DemoHelpPanel()
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /B<boardNo>    : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /T[T|E|P|4|9|G|3|6] : Board Type used for the demo, where\n");
  DemoLogf("                    T  = Use Thor-2-PCI-Plus Board\n");
  DemoLogf("                    E  = Use Thor-2-PCI-Express Board\n");
  DemoLogf("                    P  = Use Thor-8-PCI-Plus Board (default)\n");
  DemoLogf("                    4  = Use Thor-4-PCI-Plus Board\n");
  DemoLogf("                    9  = Use Gimle-16-PCIe Board\n");
  DemoLogf("                    G  = Use Gimle-16-PCI-Plus Board\n");
  DemoLogf("                    3  = Use Gimle-16-PCI-Basic_32 Board\n");
  DemoLogf("                    6  = Use Gimle-16-PCI-Basic_64 Board\n");
  DemoLogf(" /L<fileName>   : Log file name\n");
  DemoLogf(" /Z<str>        : Command string (overrides keyboard input)\n");
  DemoLogf(" /Q             : Exit without prompting on errors\n");
}


/*******************************************************************************

  @func

  Displays a message on the screen and prints it to the logfile.

  @comm

*******************************************************************************/
static void DemoLogf(char *formatStr, ...)
{
  va_list argList;
  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile,   formatStr, argList);

  va_end(argList);

}


/*******************************************************************************

  @func

  Displays a message on the screen 

  @comm

*******************************************************************************/
static void DemoPrintf(char *formatStr, ...)
{
  va_list argList;
  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);

  va_end(argList);

}

static void DemoCheckRc(IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  if ( OTX_FAILURE(result) ) {

    OTX_RESULT otxRc;
    char msgStr[256];
    
    // Convert the result code to a more descriptive string
    otxRc = OtxDrvResultCode2String(OTX_DRIVER_DEVICE_HANDLE, result, 256, &msgStr[0]);
    if (OTX_FAILURE(otxRc)) {
      strcpy(msgStr, "Cannot translate this result code.");
    }
    DemoLogf("ERROR(%#lX [%d.%d]): %s \n", OTX_RESULT_CODE(result), OTX_RESULT_SOURCE_ID(result), OTX_RESULT_LINE_NUM(result), msgStr);
    
    DemoExitQuestion();
  }
}

/*******************************************************************************
  @func
    Prints a timestamp in a MM-DD hh:mm:ss:mmm format, where

    MM  = month
    DD  = day (of month)
    hh  = hour of day
    mm  = minutes of day
    ss  = seconds of day
    mmm = milliseconds

*******************************************************************************/
void DemoPrintTimeStamp(
  OTX_DATETIME nTimeStamp
)
{
  OTX_DATETIME nTimeStampMs;
  OTX_DATETIME nTimeStampSec;
  struct Osa_tm *ptCurrentDayAndTime;
  Osa_time_t nCurrentTimeSecs;
  
  // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
  
  #if defined(__linux__)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 1164447360 * 100000000;  
  #elif defined(WIN32)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;  
  #endif

  
  // Convert to seconds since Jan 1, 1970 (so we can use this time with OsaStdLocalTime())
  nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;
  
  // Calculate number of milliseconds (since number of seconds is not a very accurate way to report timestamp)
  nTimeStampMs.nUint64 = nTimeStamp.nUint64 / 10000 - (nTimeStampSec.nUint64 * 1000);
  
  nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
  ptCurrentDayAndTime = OsaStdLocalTime(&nCurrentTimeSecs);
  
  if (ptCurrentDayAndTime) {
    
    DemoLogf("%.2d-%.2d %.2d:%.2d:%.2d.%.3d ", 
      ptCurrentDayAndTime->tm_mon + 1,
      ptCurrentDayAndTime->tm_mday,
      ptCurrentDayAndTime->tm_hour,
      ptCurrentDayAndTime->tm_min, 
      ptCurrentDayAndTime->tm_sec,
      nTimeStampMs.sUint32.nLow
      );
  } else {
    DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
  }
  
  
}

/*******************************************************************************
  @func
    Prompts whether to exit the program. Usually called after an error 
    has been reported.

*******************************************************************************/
void DemoExitQuestion()
{
  short c;

  if (g_demoEnv.bExitOnError) {
    c = 'y';
  } else {
    DemoLogf("Do you want to exit (Y/N) ");
    c = OsaConsoleGetCh();
  }
  if (c == 'y' || c == 'Y') {
    // Clean up
    OtxDrvReleaseLib();
    OsaConsoleExit();
    exit(1);
  }
}


/*******************************************************************************

  @func

  Processes User Commands

  @comm

*******************************************************************************/
OTX_INT32 DemoProcessUserCommand(char commandChar)
{
  static OTX_INT32      nCurrentDmp = 0;
  OTX_RESULT result;

  DemoLogf("%c pressed\n", commandChar);

  switch (commandChar) {

  case '?':
  case 'h':
  case 'H':
    DemoLogf("Cmd  Synopsis\n");
    DemoLogf("===  ========\n");
    DemoLogf(" ?   Display Help text\n");
    DemoLogf(" h   Ditto\n");
    DemoLogf(" q   Quit\n");
    DemoLogf(" a   Put all DMPs in Reset state\n");
    DemoLogf(" 0-3 Select current DMP. Current DMP is now DMP#%d\n", nCurrentDmp);
    DemoLogf(" b   Boot current DMP\n");
    DemoLogf(" B   Boot all DMPs\n");
    DemoLogf(" u   Set bootmode to UART for current DMP\n");
    DemoLogf(" f   Set bootmode to FLASH for current DMP\n");
    DemoLogf(" d   Set bootmode to DspSelfBoot for current DMP (Code Composer)\n");
    DemoLogf(" U   Set bootmode to UART for all DMPs\n");
    DemoLogf(" F   Set bootmode to FLASH for all DMPs\n");
    DemoLogf(" D   Set bootmode to DspSelfBoot for all DMPs (Code Composer)\n");
    break;

  case '0':
  case '1':
  case '2':
  case '3':
    // Select current DMP
    {
      OTX_INT32 nDmp = commandChar - '0';
            
      nCurrentDmp = nDmp;
      DemoLogf("LI#%d Selected ....\n", nCurrentDmp);


    }
    break;

  case 'a':
    {
      OTX_UINT32 readVal;
      OTX_UINT32 writeVal;
      static OTX_UINT32 nAddress = 0x30;

      DemoLogf("Put all DMPs in Reset state\n");

      result = OtxDrvReadReg32(g_hPhysDevs.hThorController0, nAddress, &readVal); DemoCheckRc(result);

      writeVal = readVal | 0x003C;

      result = OtxDrvWriteReg32(g_hPhysDevs.hThorController0, nAddress, writeVal); DemoCheckRc(result);

      DemoLogf("Done.\n");
    }
    break;

#define OD_ALV_DMP_BOOT_MODE      0x0008  // 3, BOOT mode (1=UART, 0=FLASH)
#define OD_ALV_DMP_DSP_BT         0x0020  // 5, DSP_BT (1=DSP boots itself (Code Composer boot config), 0=ARM boots DSP (normal operation))

#define OD_ALV_SET_REG_OFS  0x0
#define OD_ALV_CLR_REG_OFS  0x4

  case 'u':
    {
      OTX_UINT32 writeReg;

      DemoLogf("Set Boot Mode to UART for DMP#%d\n", nCurrentDmp);

      writeReg = g_demoEnv.nAsmOffset + 0x20 + (nCurrentDmp * 8);

      result = OtxDrvWriteReg32(g_hPhysDevs.hThorController0, writeReg + OD_ALV_SET_REG_OFS, OD_ALV_DMP_BOOT_MODE); DemoCheckRc(result);
      result = OtxDrvWriteReg32(g_hPhysDevs.hThorController0, writeReg + OD_ALV_CLR_REG_OFS, OD_ALV_DMP_DSP_BT); DemoCheckRc(result);

      DemoLogf("Done.\n");

    }
    break;

  case 'f':
    {
      OTX_UINT32 writeReg;

      DemoLogf("Set Boot Mode to FLASH for DMP#%d\n", nCurrentDmp);

      writeReg = g_demoEnv.nAsmOffset + 0x20 + (nCurrentDmp * 8);

      result = OtxDrvWriteReg32(g_hPhysDevs.hThorController0, writeReg + OD_ALV_CLR_REG_OFS, OD_ALV_DMP_BOOT_MODE); DemoCheckRc(result);
      result = OtxDrvWriteReg32(g_hPhysDevs.hThorController0, writeReg + OD_ALV_CLR_REG_OFS, OD_ALV_DMP_DSP_BT); DemoCheckRc(result);

      DemoLogf("Done.\n");

    }
    break;


  case 'd':
    {
      OTX_UINT32 writeReg;

      DemoLogf("Set Boot Mode to DspSelfBoot for DMP#%d\n", nCurrentDmp);

      writeReg = g_demoEnv.nAsmOffset + 0x20 + (nCurrentDmp * 8);

      result = OtxDrvWriteReg32(g_hPhysDevs.hThorController0, writeReg + OD_ALV_CLR_REG_OFS, OD_ALV_DMP_BOOT_MODE); DemoCheckRc(result);
      result = OtxDrvWriteReg32(g_hPhysDevs.hThorController0, writeReg + OD_ALV_SET_REG_OFS, OD_ALV_DMP_DSP_BT); DemoCheckRc(result);

      DemoLogf("Done.\n");

    }
    break;

  case 'U':
    {
      OTX_UINT32 writeReg;
      OTX_INT32 nDmp;

      DemoLogf("Set Boot Mode to UART for all DMPs\n");
      for (nDmp = 0; nDmp < g_demoEnv.coDmp; nDmp++) {
        
        writeReg = g_demoEnv.nAsmOffset + 0x20 + (nDmp * 8);
        
        result = OtxDrvWriteReg32(g_hPhysDevs.hThorController0, writeReg + OD_ALV_SET_REG_OFS, OD_ALV_DMP_BOOT_MODE); DemoCheckRc(result);
        result = OtxDrvWriteReg32(g_hPhysDevs.hThorController0, writeReg + OD_ALV_CLR_REG_OFS, OD_ALV_DMP_DSP_BT); DemoCheckRc(result);
      }
      DemoLogf("Done.\n");

    }
    break;


  case 'F':
    {
      OTX_UINT32 writeReg;
      OTX_INT32 nDmp;

      DemoLogf("Set Boot Mode to FLASH for all DMPs\n");
      for (nDmp = 0; nDmp < g_demoEnv.coDmp; nDmp++) {
        
        writeReg = g_demoEnv.nAsmOffset + 0x20 + (nDmp * 8);
        
        result = OtxDrvWriteReg32(g_hPhysDevs.hThorController0, writeReg + OD_ALV_CLR_REG_OFS, OD_ALV_DMP_BOOT_MODE); DemoCheckRc(result);
        result = OtxDrvWriteReg32(g_hPhysDevs.hThorController0, writeReg + OD_ALV_CLR_REG_OFS, OD_ALV_DMP_DSP_BT); DemoCheckRc(result);
      }
      DemoLogf("Done.\n");

    }
    break;


  case 'D':
    {
      OTX_UINT32 writeReg;
      OTX_INT32 nDmp;

      DemoLogf("Set Boot Mode to DspSelfBoot for all DMPs\n");
      for (nDmp = 0; nDmp < g_demoEnv.coDmp; nDmp++) {
        
        writeReg = g_demoEnv.nAsmOffset + 0x20 + (nDmp * 8);
        
        result = OtxDrvWriteReg32(g_hPhysDevs.hThorController0, writeReg + OD_ALV_CLR_REG_OFS, OD_ALV_DMP_BOOT_MODE); DemoCheckRc(result);
        result = OtxDrvWriteReg32(g_hPhysDevs.hThorController0, writeReg + OD_ALV_SET_REG_OFS, OD_ALV_DMP_DSP_BT); DemoCheckRc(result);
      }
      DemoLogf("Done.\n");

    }
    break;


  case 'b':
    {
      OTX_UINT32 readVal;
      OTX_UINT32 writeVal;
      static OTX_UINT32 nAddress = 0x30;

      DemoLogf("Boot DMP#%d\n", nCurrentDmp);

      result = OtxDrvReadReg32(g_hPhysDevs.hThorController0, nAddress, &readVal); DemoCheckRc(result);

      writeVal = readVal | (0x0001 << (nCurrentDmp + 2));

      result = OtxDrvWriteReg32(g_hPhysDevs.hThorController0, nAddress, writeVal); DemoCheckRc(result);

      OsaThreadSleep(50);

      writeVal = readVal & ~(0x0001 << (nCurrentDmp + 2));

      result = OtxDrvWriteReg32(g_hPhysDevs.hThorController0, nAddress, writeVal); DemoCheckRc(result);

      DemoLogf("Done.\n");

    }
    break;


  case 'B':
    {
      OTX_UINT32 readVal;
      OTX_UINT32 writeVal;
      static OTX_UINT32 nAddress = 0x30;

      DemoLogf("Boot all DMPs\n");

      result = OtxDrvReadReg32(g_hPhysDevs.hThorController0, nAddress, &readVal); DemoCheckRc(result);

      writeVal = readVal | 0x003C;

      result = OtxDrvWriteReg32(g_hPhysDevs.hThorController0, nAddress, writeVal); DemoCheckRc(result);

      OsaThreadSleep(50);

      writeVal = readVal & ~0x003C;

      result = OtxDrvWriteReg32(g_hPhysDevs.hThorController0, nAddress, writeVal); DemoCheckRc(result);

      DemoLogf("Done.\n");

    }
    break;



  case '~':
    {
      char strBuf[12];
      OTX_INT32 nSleepTime;

      DemoPrintf("Enter milliseconds to sleep :");
      OsaConsoleGetStr(strBuf);

      if (strBuf[0]) {
        nSleepTime = strtol(strBuf, NULL, 0);
        DemoPrintf("Sleeping %d milliseconds\n", nSleepTime);
        OsaThreadSleep(nSleepTime);
      }
    }
    break;


  case 'q':
  case 'Q':
    return 1;
    break;

  default:
    DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
    break;
  }

  return 0;
}


/*******************************************************************************
*******************************************************************************/
void DemoAtExit(void)
{
  fclose(g_pDemoLogFile);
}


/*******************************************************************************
*******************************************************************************/
int main(int argc, char *argv[])
{
  OTX_INT32 exitFlag = 0;
  OTX_RESULT result;  // OTX Standard Return Code
  char commandChar;

  OsaConsoleInit();

  DemoLogf("ALVISBOOT, %s, Odin TeleSystems Inc.\n", __DATE__);
  DemoLogf("Board Device Access Example program.\n");
  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if (g_demoEnv.pszInputStr) {
    OsaConsoleInput(g_demoEnv.pszInputStr);
  }

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
    OsaConsoleExit();
		exit(-1);
  }
  atexit(DemoAtExit);

  DemoLogf("Log file is  %s\n", g_demoEnv.logFileName);
  DemoLogf("Board number %d\n", g_demoEnv.boardNo);


  // Query Library revision
  {
    char szVersionInfo[255];
    OTX_BOOL bIsStaticLinkedLibrary;
    char szLibFileName[64];
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_REV_STR, szVersionInfo, 255); DemoCheckRc(result);
    result = OtxDrvGetAttributeValueBool(OTX_LIB_HANDLE, OTX_ATTR_LIB_TYPE_STATIC, &bIsStaticLinkedLibrary); DemoCheckRc(result);
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_FILE_NAME, szLibFileName, 64); DemoCheckRc(result);
    
    DemoLogf("Library version %s (filename %s, %s linked)\n", szVersionInfo, szLibFileName, bIsStaticLinkedLibrary ? "static":"dynamic");
  }

  // Connect to the Driver
  DemoLogf("Connecting lib...\n");
  result = OtxDrvConnectLib(); 
  switch (OTX_RESULT_CODE(result)) {
  case OTX_RES_HANDLE:
    DemoLogf("Error: Could not connect to the OTX library.\nPossible cause: OTX Driver not loaded\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  case OTX_RES_NOT_SUPPORTED:
    DemoLogf("Error: Mismatch between OTX Driver revision and OTX Library revision.\nRevisions must match.\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  default:
    DemoCheckRc(result);
    break;
  }

  // Query driver revision
  {
    char szVersionInfo[255];
    result = OtxDrvGetAttributeValueSz(OTX_DRIVER_DEVICE_HANDLE, OTX_ATTR_DRV_REV_STR, szVersionInfo, 255); DemoCheckRc(result);
    
    DemoLogf("Driver version %s\n", szVersionInfo);
  }

  // Connect (get handles) to the physical devices on the Thor-2-PCI board
  DemoLogf("Get Handle to the board device...\n");
  DemoOpenPhysicalDevices();


  // Read Serial Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszSerialNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_SERIAL_NUMBER, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszSerialNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_SERIAL_NUMBER, pszSerialNumber, cbAttribValue); DemoCheckRc(result);
    DemoLogf("Board Serial Number: %s\n", pszSerialNumber);
    free(pszSerialNumber);
  }

  // Read Build Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszBrdBuildNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszBrdBuildNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_BUILD, pszBrdBuildNumber, cbAttribValue); DemoCheckRc(result);
    DemoLogf("Board Build Number: %s\n", pszBrdBuildNumber);
    free(pszBrdBuildNumber);
  }

  if (g_demoEnv.nBoardType != OTX_DEVICE_THOR_PCMCIA) {
    // Read CPLD Build Numbers from the board
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszCpldBuildNumbers;
    
    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_CPLD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszCpldBuildNumbers = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_CPLD_BUILD, pszCpldBuildNumbers, cbAttribValue); DemoCheckRc(result);
    DemoLogf("CPLD Build:\n%s", pszCpldBuildNumbers);
    free(pszCpldBuildNumbers);
  }

  {
    OTX_UINT32 nSubsystemId;
    OTX_UINT32 nDeviceId;
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_DEVICE_ID, &nDeviceId); DemoCheckRc(result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_SUBSYSTEM_ID,  &nSubsystemId); DemoCheckRc(result);
    DemoLogf("Board Device ID: 0x%.4X, SubsystemId: 0x%.4X\n", nDeviceId, nSubsystemId);
    if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI_PLUS) {
      if (nSubsystemId == OTX_NIC_SUBSYSTEM_ID_REV_2_0) {
        g_demoEnv.nAsmOffset    = 0x100; 
      } else {
        g_demoEnv.nAsmOffset    = 0x80;  
      }
    }
  }


  // Main Loop. Loop here and wait for Events from the driver or commands from the user ....

  DemoLogf("\nPress 'h' for help\n");

  while (!exitFlag) {
    commandChar = OsaConsoleGetCh();
    exitFlag = DemoProcessUserCommand(commandChar);
  }

  // Clean up
  OtxDrvReleaseLib();
  OsaConsoleExit();

  return 0;
}


#if defined(_WIN32_WCE)
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPWSTR lpCmdLine, int nShowCmd) 
{
  // default command line parameters
  static char exename[OTX_MAX_PATH * sizeof(TCHAR)] = "";
  static char *cmdParams[] = { exename, "/TM", "/MT", "/D4", NULL }; 
  char *pc;

  GetModuleFileName(NULL, (LPTSTR) cmdParams[0], OTX_MAX_PATH);
  OsaStrCpy(g_demoEnv.cwd, cmdParams[0]);
  pc = OsaStrRChr(g_demoEnv.cwd, '\\');
  if (pc != NULL) {
    *++pc = '\0';
  }

  return OsaWinCeConsoleMain(hInstance, hPrevInstance, lpCmdLine, nShowCmd,
              cmdParams, sizeof(cmdParams) / sizeof(cmdParams[0]), main);
}
#endif  // #if defined(_WIN32_WCE)
