/* $Workfile: T8pBurstBasic.c $       $Revision: 28 $      $Modtime: 10/30/09 11:23a $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    T8pBurstBasic.c

    A simple example console application demonstrating the use of the
    DMA Direct Sending and Receiving with the OTX Driver using the
    board's Burst device to access the data using DMA.

      Thor-2-PCI-Plus       (burst data looped via TSS)
      Thor-2-PCI-Express    (burst data looped via TSS)
      Thor-4-PCI-Plus       (burst data looped via TSS)
      Thor-8-PCI-Plus       (burst data looped via TSS)
      Thor-2-ExpressCard    (burst data looped via E1T1 devices and the DSP device)
      Thor-4-ExpressCard    (burst data looped via E1T1 devices and the DSP device)

    The Demo application provides a simple console interface which utilizes
    simple keystrokes as commands.
    
    It configures the burst device which is looped back to the burst device via
    the time-space-switch (if available) or via T1E1 devices.

    The direct sends and receives operate on a highway basis and are
    hosted by the Burst device.

    Press 'h' or '?' for help. 'q' to quit.

*******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <assert.h>

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxErr.h"
#include "OtxDef.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxBrd.h"
#include "OtxT1E1.h"  
#include "OtxDsp.h"  

#include "OtxBurst.h"
#include "OtxBurstD.h"

#include "Dsp55Demo2Io.h"   // IO Control codes for the DSP to host communication
#include <Dsp55Ioctl.h>     // IO Control codes for the DSP to host communication


#define DEMO_DEFAULT_LOG_FILE   "T8pBurstBasic.log"
#define DEMO_DEFAULT_BOARD_NO   0
#define DEMO_MAX_DSP            2

#define DEMO_BURST_BUF_SIZE     0x8000    // 32k


typedef struct DemoEnvT {
  char                  *pszInputStr;
  OTX_BOOL               bExitOnError;
  char                   cwd[OTX_MAX_PATH];
  char                   logFileName[OTX_MAX_PATH];
  OtxPhysicalDeviceTypeE nBoardType;
  int                    coBurstHighways;
  int                    nBurstHwyTimeslots;
  int                    nCurrentHwy;
  int                    boardNo;
  int                    nDspNo;
  int                    coLi;
  OTX_BOOL               bDisableBurstTx;
  OTX_BOOL               bDisableBurstRx;
  OTX_BOOL               bUseExternalLoop;
  OTX_BOOL               bUseE1Loop;
  OTX_BOOL               bUseDspLoop;
  OTX_BOOL               bUseCodecLoop;
  OTX_BOOL               bUseH100Loop;
  OTX_INT32              nHbusRate;
  OTX_INT32              nH100HwyRate;
  OtxDataFormatE         eLaw;
  OtxBrdClockSourceE     eClkSource;
} DemoEnvT;


/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoCreateEventQueue(void);
static void DemoCreatePhysicalDevices(void);
static void DemoConfigureE1T1Devices(void);
static void DemoEnablePhysicalDevices(void);
static void DemoConnectDevices(void);
static void DemoPrintEvent(IN OtxEventDataS *pEventData);
static void DemoCheckRc(IN OTX_RESULT otxRc);
static void DemoExitQuestion(void);
static void DemoLogf(char *formatStr, ...);
static void DemoHelpPanel(void);
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);
static void DemoBurstSender(OTX_UINT32 nSourceId);
static void DemoBurstReceiver(OTX_UINT32 nSourceId);


/*******************************************************************************
 Global Data
*******************************************************************************/

OTX_EVENT   g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hEventQueue  = 0;
FILE        *g_pDemoLogFile = NULL;
DemoEnvT    g_demoEnv;
OTX_BOOL    g_bStatsThreadActive = OTX_FALSE;
OTX_BOOL    g_bReadBurstStatsThreadActive = OTX_FALSE;
OTX_UINT32  g_nDspHighwaySpeed;  // DSP Highway speed in kpbs
OTX_HANDLE  g_hUserApp[DEMO_MAX_DSP];
OTX_HANDLE  g_hDsp[DEMO_MAX_DSP];
OTX_HANDLE  g_hAsm = OTX_INVALID_HANDLE_VALUE;
OTX_INT32   g_exitFlag = 0;
OTX_INT32   g_nExitCode = 0;
OTX_INT32   g_nOverRunErrors = 0;
OTX_INT32   g_nUnderRunErrors = 0;
OTX_INT32   g_nNoRxDataErrors = 0;

struct {
  OTX_HANDLE hThorController;
  OTX_HANDLE hBurst;
  OTX_HANDLE hTss;
  OTX_HANDLE hLi[8];
  OTX_HANDLE hCodec[4];  // Just used if g_demoEnv.bUseCodecLoop is OTX_TRUE
  OTX_HANDLE hCtBus;
} g_hPhysDevs;

// direct Burst DMA send
struct sBurstSender {
  OTX_UINT32  nSourceId;
  OTX_BOOL    bVerbose;
  OTX_BOOL    bActive;
  OTX_UINT8   sqTxBuf[DEMO_BURST_BUF_SIZE];
  FILE        *pTxFile;
  OTX_INT32   nTsToSend;  // Default is -1 (All TS)
  OTX_UINT32  cbTxBuf;
  OTX_BOOL    bFileLoop;

  OTX_UINT8   sqGammaData[128];

};

// direct Burst DMA receive
struct sBurstReceiver {
  OTX_UINT32  nSourceId;
  OTX_BOOL    bVerbose;
  OTX_BOOL    bActive;
  FILE        *pRxFile;
  OTX_INT32   nTsToRecord;  // Default is -1 (All TS)
  OTX_UINT32  nRecvCnt;

  OTX_INT32   nDataBlockCounter;
  OTX_BOOL    bGammaTestMode;
  OTX_INT32   nGammaTestErrors[128];
  OTX_UINT8   sqGammaTestData[128];
};

struct {
  struct sBurstSender mBurstSender[2];        // one for each possible highway hosted by the Burst device
  struct sBurstReceiver mBurstReceiver[2];    // one for each possible highway hosted by the Burst device
} g_hLogicDevs;



/*******************************************************************************
*******************************************************************************/


void DemoCreateEventQueue(void)
/*******************************************************************************

  @func

  Create the event queue to be used by the driver to inform the application
  about events, such as line status changes, or detected DMTF digits.

*******************************************************************************/
{
  OTX_RESULT result;

  // Create a notification Event
  result = OtxDrvCreateEvent(&g_hNotificationEvent);
  if (OTX_RESULT_CODE(result) == OTX_E_NOTIMPL) {
    g_hNotificationEvent = NULL;
    result = OTX_S_OK;
  }

  DemoCheckRc(result);

  // Create Event Queue for driver notifications
  result = OtxDrvCreateEventQueue(128, g_hNotificationEvent, &g_hEventQueue); DemoCheckRc(result);

  // Enable Event Queue
  result = OtxDrvEnable(g_hEventQueue, OTX_TASK_SYNC, 5000L); DemoCheckRc(result);

}



void DemoCreatePhysicalDevices(void)
/*******************************************************************************

  @func

  Create and Connect the Physical Devices on the Thor-8 board

  @comm

    Calls function <f OtxDrvCreatePhysicalDevice> returning a handle to a
    Device Object representing a physical device, which can be the
    entire board or an integrated circuit (chip) on the board. The number of
    physical devices is set per board and cannot change dynamically.

    The physical devices can be classified in 2 categories:

  - BOARD Devices: A board device represents one OTX Network Interface card or
    a Daughter Board Module. The board device is the place holder and controller
    of other physical and sometime logical devices representing
    the devices on the board: such as Line Interface and Filter  Devices.

    The Burst device hosts the DMA functionality.  If we open the Burst
    device, when we enable that device, the process of transferring data from
    the board to the driver will take place via DMA.  The Burst device provides
    the necessary buffering to process the data from the board until it can be
    processed and filtered to the logical devices that are hosted by Burst.
    (see DemoCreateLogicalDevices for example of connection).

  - CHIP Devices: A chip device represents a physical integrated circuit
    on the board performing one or several functions. A logical device can
    be created to represent each of the functions implemented by a chip device.
    For example, a Digital Signal Processor is a physical "chip" device.
    However, the DSP software may implement several functions and be
    represented by a multiple of logical devices.



*******************************************************************************/
{
  OTX_RESULT result;

  // Connect to the Physical Board Controller
  switch (g_demoEnv.nBoardType) {
    case OTX_DEVICE_THOR_2_PCI_PLUS:
      DemoLogf("Opening Thor-2-PCI-Plus board\n");
      break;
    case OTX_DEVICE_THOR_2_PCI_EXPRESS:
      DemoLogf("Opening Thor-2-PCI-Express board\n");
      break;
    case OTX_DEVICE_THOR_4_PCI_PLUS:
      DemoLogf("Opening Thor-4-PCI-Plus board\n");
      break;
    case OTX_DEVICE_THOR_8_PCI_PLUS:
      DemoLogf("Opening Thor-8-PCI-Plus board\n");
      break;
    case OTX_DEVICE_THOR_1_PCI_PLUS:
      DemoLogf("Opening Thor-1-PCI-Plus board\n");
      break;
    case OTX_DEVICE_THOR_2_EXPRESSCARD:
      DemoLogf("Opening Thor-2-ExpressCard board\n");
      break;
    case OTX_DEVICE_THOR_4_EXPRESSCARD:
      DemoLogf("Opening Thor-4-ExpressCard board\n");
      break;
    case OTX_DEVICE_HUGIN_2_PC104:
      DemoLogf("Opening Hugin-2-PC104 board\n");
      break;
    default:
      DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
      OsaConsoleExit();
      exit(1);
      break;
  }

  result = OtxDrvOpenPhysicalDevice(0, g_demoEnv.nBoardType, g_demoEnv.boardNo, g_demoEnv.boardNo, g_hEventQueue, &g_hPhysDevs.hThorController); DemoCheckRc(result);

  // Get a handle to the physical Burst device.
  result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController, OTX_DEVICE_BURST, 0, 0, g_hEventQueue, &g_hPhysDevs.hBurst); DemoCheckRc(result);

  // set device buffer event trigger levels
  result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_RECEIVER_FIFO_TRIGGER,  DEMO_BURST_BUF_SIZE); DemoCheckRc(result);
  result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_SENDER_FIFO_TRIGGER,    DEMO_BURST_BUF_SIZE); DemoCheckRc(result);

  // Optionally (mainly for testing, or for improving system performance for single direction burst applications,
  // disable one direction of the burst device
  if (g_demoEnv.bDisableBurstRx) {
    DemoLogf("Burst RX interrupts will be turned off when the Burst device is enabled\n");
    result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hBurst, OTX_ATTR_BURST_RECV_DISABLED,  OTX_TRUE); DemoCheckRc(result);
  }
  if (g_demoEnv.bDisableBurstTx) {
    DemoLogf("Burst TX interrupts will be turned off when the Burst device is enabled\n");
    result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hBurst, OTX_ATTR_BURST_SEND_DISABLED,  OTX_TRUE); DemoCheckRc(result);
  }
}




void DemoEnablePhysicalDevices(void)
/*******************************************************************************

  @func
    Enable Configured Physical Devices

*******************************************************************************/
{
  OTX_RESULT result;

  // Enable the board (Physical Device)
  DemoLogf("Enable g_hPhysDevs.hThorController\n");
  result = OtxDrvEnable(g_hPhysDevs.hThorController,  OTX_TASK_SYNC, 5000L); DemoCheckRc(result);

  DemoLogf("Enable g_hPhysDevs.hBurst\n");
  result = OtxDrvEnable(g_hPhysDevs.hBurst, OTX_TASK_SYNC, 5000L);

}


/*******************************************************************************

  @func

*******************************************************************************/
void DemoPrintPipeConnections(
  OTX_HANDLE hPipe
)
{
  OTX_RESULT result;
  OTX_UINT32 nCapacity;
  OTX_UINT32 nInHandle;
  OTX_UINT32 nOutHandle;
  OTX_UINT32 nInHighway;
  OTX_UINT32 nOutHighway;
  OTX_UINT32 cbReturned;
  OTX_UINT8 sqInTimeSlotMask[128];
  OTX_UINT8 sqOutTimeSlotMask[128];
  OTX_UINT32 n;
    
  result = OtxDrvGetAttributeValueUint32(hPipe, OTX_ATTR_PIPE_CAPACITY, &nCapacity); 
  DemoCheckRc(result);

  result = OtxDrvGetAttributeValueUint32(hPipe, OTX_ATTR_PIPE_IN_CONNECTED, &nInHandle); 
  DemoCheckRc(result);
  result = OtxDrvGetAttributeValueUint32(hPipe, OTX_ATTR_PIPE_OUT_CONNECTED, &nOutHandle); 
  DemoCheckRc(result);

  result = OtxDrvGetAttributeValueUint32(hPipe, OTX_ATTR_PIPE_IN_HIGHWAY, &nInHighway); 
  DemoCheckRc(result);
  result = OtxDrvGetAttributeValueUint32(hPipe, OTX_ATTR_PIPE_OUT_HIGHWAY, &nOutHighway); 
  DemoCheckRc(result);

  result = OtxDrvGetAttributeValue(hPipe, OTX_ATTR_PIPE_IN_CONN_MASK, sqInTimeSlotMask, 128, &cbReturned); 
  DemoCheckRc(result);
  result = OtxDrvGetAttributeValue(hPipe, OTX_ATTR_PIPE_OUT_CONN_MASK, sqOutTimeSlotMask, 128, &cbReturned); 
  DemoCheckRc(result);

  DemoLogf("Connection info for pipe %X (Capacity:%d)\n", hPipe, nCapacity);

  DemoLogf("  Local input  connections: Hwy:%d ", nInHighway);

  DemoLogf("TS: ");
  for (n = 0; n < 128; n++) {
    if (sqInTimeSlotMask[n]) {
      DemoLogf("%d ", n);
    }
  }
  DemoLogf("\n");

  DemoLogf("  Local output connections: Hwy:%d ", nOutHighway);

  DemoLogf("TS: ");
  for (n = 0; n < 128; n++) {
    if (sqOutTimeSlotMask[n]) {
      DemoLogf("%d ", n);
    }
  }
  DemoLogf("\n");

  // CT Bus connection info
/*
  {
    OtxPipeCtBusConnectionS *psqCtBusConnectPoints;
    OTX_UINT32  cbCtBusConnectPoints;  
    OTX_UINT32  coCtBusConnectPoints;  
    OTX_UINT32  cbAttribValue;   // Bytes actually used in psqCtBusConnectPoints

    result = OtxDrvGetAttributeValueUint32(hPipe, OTX_ATTR_PIPE_CT_BUS_CONNECTIONS_COUNT, &coCtBusConnectPoints); DemoCheckRc(result);
    cbCtBusConnectPoints = (coCtBusConnectPoints * sizeof(OtxPipeCtBusConnectionS));

    psqCtBusConnectPoints = (OtxPipeCtBusConnectionS *) malloc(cbCtBusConnectPoints * sizeof(char));
    assert(psqCtBusConnectPoints);

    result = OtxDrvGetAttributeValue(hPipe, OTX_ATTR_PIPE_CT_BUS_CONNECTIONS_DETAILS, psqCtBusConnectPoints, cbCtBusConnectPoints, &cbAttribValue); DemoCheckRc(result);
    assert(cbAttribValue == cbCtBusConnectPoints);

    if (coCtBusConnectPoints) {
      for (n = 0; n < coCtBusConnectPoints; n++) {
        DemoLogf("  Sourcing CT Bus:%d, CT Highway:%d, CT TS:%d\n", psqCtBusConnectPoints[n].nCtBus, psqCtBusConnectPoints[n].nHighway, psqCtBusConnectPoints[n].nTimeslot );
      }
    } else {
      DemoLogf("  No sourced CT Bus connections\n");
    }

    free(psqCtBusConnectPoints);
  }
*/
    
}


void DemoConnectH100(
/*******************************************************************************

  @func
    Create the needed Pipe Devices and connect the physical and logical devices
    to pipes.

  @comm
    A Pipe Device is a logical entity representing a serial bit-stream (a subset 
    of a Highway) with a data rate between 8 kbit/s and 2.048Mbit/s. A pipe is a 
    characterized but its capacity (in 8 kbit/s increments). The user can create 
    a pipe with the driver function <f OtxDrvCreatePipe>. A pipe is connected to 
    physical or logical devices and mapped to specific time-slots with the driver 
    function <f OtxDrvConnectPipe>.

*******************************************************************************/
  OTX_UINT32  nTransferType
)
{
  OTX_RESULT result;
  OTX_HANDLE hPipe;
  OTX_INT32  nTs;
  OTX_INT32  nH100Ts;
  OTX_INT32  nH100Highway;
  OTX_UINT8  sqBurstTimeSlotMask[OTX_PIPE_TIMESLOTS_MAX];
  OTX_UINT8  sqH100TimeSlotMask[OTX_PIPE_TIMESLOTS_MAX];
  OTX_INT32  sqTssBurstHighway[4];
  OTX_INT32  sqTssH100Highway[4];
  OTX_UINT32 nH100MaskSize;
  OTX_UINT8  *pH100Mask;

  sqTssH100Highway[0]  = 16; 
  sqTssH100Highway[1]  = 17; 
  sqTssH100Highway[2]  = 18; 
  sqTssH100Highway[3]  = 19; 

  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_2_PCI_PLUS:
  case OTX_DEVICE_THOR_2_PCI_EXPRESS:
    sqTssBurstHighway[0]  = 4;
    break;
  case OTX_DEVICE_THOR_4_PCI_PLUS:
    sqTssBurstHighway[0]  = 14;
    break;
  case OTX_DEVICE_THOR_8_PCI_PLUS:
    sqTssBurstHighway[0]  = 14;
    break;
  case OTX_DEVICE_THOR_1_PCI_PLUS:
    sqTssBurstHighway[0]  = 4;
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    OsaConsoleExit();
    exit(1);
    break;
  }

  DemoLogf("Creating pipes for Burst %s TssH100Hwy=%d, TssBurstHwy=%d, nBurstHwyTimeslots=%d\n", 
           nTransferType == 0 ? "Sender" : "Receiver",
           sqTssH100Highway[0], 
           sqTssBurstHighway[0], 
           g_demoEnv.nBurstHwyTimeslots);

  nH100Highway = 0;
  nH100Ts = 0;
  // --------------------------------------------------------------------
  // Create pipe: Burst -> H100, one pipe for each time slot
  // --------------------------------------------------------------------
  for (nTs = 0; nTs < g_demoEnv.nBurstHwyTimeslots; nTs++) {

    result = OtxDrvCreatePipe(64, &hPipe); DemoCheckRc(result);

    memset(sqBurstTimeSlotMask, 0, sizeof(sqBurstTimeSlotMask));
    memset(sqH100TimeSlotMask, 0, sizeof(sqH100TimeSlotMask));
    sqBurstTimeSlotMask[nTs]    = 0xFF;
    sqH100TimeSlotMask[nH100Ts] = 0xFF;

    // The Highway is configured by the driver during initialization
    // and may not support the selected data rate.
    if(g_demoEnv.nH100HwyRate == 2048) {
      pH100Mask = sqH100TimeSlotMask;
      nH100MaskSize = 32;
    }
    else
    if(g_demoEnv.nH100HwyRate == 8192) {
      pH100Mask = sqBurstTimeSlotMask;
      nH100MaskSize = 128;
    }
    else {
      // Not supported at this time
      DemoLogf("Creating pipes for Burst size %d not supported\n", g_demoEnv.nH100HwyRate);
      return;
    }


    if(nTransferType == 0) {
      // ---------------------------
      // Connect pipe: Burst -> H100
      // ---------------------------
      result = OtxDrvConnectPipe(hPipe, 
                                 OTX_PIPE_INPUT,  
                                 g_hPhysDevs.hTss, 
                                 sqTssBurstHighway[0], 
                                 sqBurstTimeSlotMask, 
                                 g_demoEnv.nBurstHwyTimeslots);
      DemoCheckRc(result);
    
      result = OtxDrvConnectPipe(hPipe, 
                                 OTX_PIPE_OUTPUT, 
                                 g_hPhysDevs.hTss, 
                                 sqTssH100Highway[nH100Highway], 
                                 pH100Mask, 
                                 nH100MaskSize);
      DemoCheckRc(result);
    }
    else 
    if(nTransferType == 1) {
      // --------------------------- 
      // Create pipe: H100 -> Burst
      // --------------------------- 
      result = OtxDrvCreatePipe(64, &hPipe); DemoCheckRc(result);
    
      // ---------------------------
      // Connect pipe: H100 -> Burst
      // ---------------------------
      result = OtxDrvConnectPipe(hPipe, 
                                 OTX_PIPE_INPUT,  
                                 g_hPhysDevs.hTss, 
                                 sqTssH100Highway[nH100Highway], 
                                 pH100Mask, 
                                 nH100MaskSize);
      DemoCheckRc(result);
    
      result = OtxDrvConnectPipe(hPipe, 
                                 OTX_PIPE_OUTPUT, 
                                 g_hPhysDevs.hTss, 
                                 sqTssBurstHighway[0], 
                                 sqBurstTimeSlotMask, 
                                 g_demoEnv.nBurstHwyTimeslots);
      DemoCheckRc(result);
    }
    else {
      DemoLogf("Type (%d) not supported by this demo program\n", nTransferType);
      nTs = g_demoEnv.nBurstHwyTimeslots; // exit for loop
    }

    //DemoPrintPipeConnections(hPipe);

    nH100Ts++;
    if((nH100Ts == 32) && (nTs < g_demoEnv.nBurstHwyTimeslots - 1)) {
      nH100Ts = 0;
      nH100Highway++;
      DemoLogf("Creating pipes for Burst %s TssH100Hwy=%d, TssBurstHwy=%d, BurstTs=%d\n", 
               nTransferType == 0 ? "Sender" : "Receiver",
               sqTssH100Highway[nH100Highway], 
               sqTssBurstHighway[0],
               nTs + 1);
    }
  }
}

void DemoConnectDevices(
/*******************************************************************************

  @func
    Create the needed Pipe Devices and connect the physical and logical devices
    to pipes.

  @comm
    A Pipe Device is a logical entity representing a serial bit-stream (a subset 
    of a Highway) with a data rate between 8 kbit/s and 2.048Mbit/s. A pipe is a 
    characterized but its capacity (in 8 kbit/s increments). The user can create 
    a pipe with the driver function <f OtxDrvCreatePipe>. A pipe is connected to 
    physical or logical devices and mapped to specific time-slots with the driver 
    function <f OtxDrvConnectPipe>.

*******************************************************************************/
  void
)
{
  OTX_RESULT result;

  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_2_PCI_PLUS:
  case OTX_DEVICE_THOR_2_PCI_EXPRESS:
  case OTX_DEVICE_THOR_4_PCI_PLUS:
  case OTX_DEVICE_THOR_8_PCI_PLUS:
  case OTX_DEVICE_THOR_1_PCI_PLUS:
    if (g_demoEnv.bUseCodecLoop) {
      // Loop via the Codecs. 4 timeslots. Requires an external Codec loop plug. 
      OTX_INT32 iHighwayNo;
      OTX_HANDLE hPipe;
      OTX_UINT8 sqBurstTimeSlotMask[OTX_PIPE_TIMESLOTS_MAX];
      OTX_INT32 nCodecNo;
      
      // Create the Pipes (one for each direction) between the Burst Device and the Codec
      iHighwayNo = 0;

      for (nCodecNo = 0 ; nCodecNo < 4; nCodecNo++) {

        // Open Codec device
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController, OTX_DEVICE_CODEC, nCodecNo, nCodecNo, g_hEventQueue, &g_hPhysDevs.hCodec[nCodecNo]); DemoCheckRc(result);

        // Set PCM compression law for the Codec
        DemoLogf("Set Law to %s\n", g_demoEnv.eLaw == OTX_DATA_FORMAT_COMPAND8_A_LAW ? "A-law" : "U-law");
        result = OtxBrdCodecSetLaw(g_hPhysDevs.hCodec[nCodecNo], g_demoEnv.eLaw); DemoCheckRc(result);
      }
        
      for (nCodecNo = 0 ; nCodecNo < 4; nCodecNo++) {
        memset(sqBurstTimeSlotMask, 0x00, g_demoEnv.nBurstHwyTimeslots);
        sqBurstTimeSlotMask[nCodecNo] = 0xFF;
        
        DemoLogf("Creating pipe for Burst Sender -> Codec[%d] [burstHighway=%d, burstTS=%d]\n", nCodecNo, iHighwayNo, nCodecNo);
        result = OtxDrvCreatePipe(64, &hPipe); DemoCheckRc(result);
        
        // -------------------------
        // Connect pipe: BurstSender -> Codec speaker
        // -------------------------
        result = OtxDrvConnectPipe(hPipe, OTX_PIPE_INPUT,  g_hPhysDevs.hBurst, iHighwayNo, sqBurstTimeSlotMask, g_demoEnv.nBurstHwyTimeslots);
        DemoCheckRc(result);
        
        result = OtxDrvConnectPipe(hPipe, OTX_PIPE_OUTPUT, g_hPhysDevs.hCodec[nCodecNo], -1, NULL, 0);
        DemoCheckRc(result);
        
        DemoLogf("Creating pipe for Codec[%d] -> Burst Receiver [burstHighway=%d, burstTS=%d]\n", nCodecNo, iHighwayNo, nCodecNo);
        result = OtxDrvCreatePipe(64, &hPipe); DemoCheckRc(result);
        
        // -------------------------
        // Connect pipe:  Codec mic -> BurstReceiver
        // -------------------------
        result = OtxDrvConnectPipe(hPipe, OTX_PIPE_INPUT, g_hPhysDevs.hCodec[nCodecNo], -1, NULL, 0);
        DemoCheckRc(result);

        result = OtxDrvConnectPipe(hPipe, OTX_PIPE_OUTPUT,  g_hPhysDevs.hBurst, iHighwayNo, sqBurstTimeSlotMask, g_demoEnv.nBurstHwyTimeslots);
        DemoCheckRc(result);
        
      }
      DemoLogf("Codec pipe connections are established\n");

    } else if (g_demoEnv.bUseExternalLoop) {
      OTX_INT32 iHighwayNo;
      OTX_HANDLE hPipe;
      OTX_UINT8 sqBurstTimeSlotMask[OTX_PIPE_TIMESLOTS_MAX];
      OTX_INT32 sqTssBurstHighway[4];
      OTX_INT32 sqTssLiHighway[4];
      OTX_INT32 nLi;
      OTX_INT32 nQuadE1T1;
      OTX_INT32 coLi = 0;
      OTX_INT32 coQuadE1T1 = 0;
      OTX_HANDLE hQuadE1T1[2];
      OTX_HANDLE hLi[8];
      OtxE1ConfigOptionsS sE1CfgData;
      
      switch (g_demoEnv.nBoardType) {
      case OTX_DEVICE_THOR_2_PCI_PLUS:
      case OTX_DEVICE_THOR_2_PCI_EXPRESS:
        sqTssLiHighway[0]     = 0;
        sqTssLiHighway[1]     = 1;
        sqTssBurstHighway[0]  = 4;
        sqTssBurstHighway[1]  = 5;
        coLi = 2;
        coQuadE1T1 = 0;
        break;
      case OTX_DEVICE_THOR_4_PCI_PLUS:
        sqTssLiHighway[0]     = 12;
        sqTssBurstHighway[0]  = 14;
        coLi = 4;
        coQuadE1T1 = 1;
        break;
      case OTX_DEVICE_THOR_8_PCI_PLUS:
        sqTssLiHighway[0]     = 12;
        sqTssLiHighway[1]     = 13;
        sqTssBurstHighway[0]  = 14;
        sqTssBurstHighway[1]  = 15;
        coLi = 8;
        coQuadE1T1 = 2;
        break;
      case OTX_DEVICE_THOR_1_PCI_PLUS:
        sqTssLiHighway[0]     = 0;
        sqTssBurstHighway[0]  = 4;
        coLi = 1;
        coQuadE1T1 = 0;
        break;
      default:
        DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
        OsaConsoleExit();
        exit(1);
        break;
      }
      
      // Create the Pipes (one for each direction) between the Burst Device and the LIs
      for (iHighwayNo = 0; iHighwayNo < g_demoEnv.coBurstHighways; iHighwayNo++) {
        OTX_INT32 nTs;
        
        DemoLogf("Creating pipes for Burst Sender/Receiver highway=%d TssLiHwy=%d, TssBurstHwy=%d, nBurstHwyTimeslots=%d\n", iHighwayNo, sqTssLiHighway[iHighwayNo], sqTssBurstHighway[iHighwayNo], g_demoEnv.nBurstHwyTimeslots);
        
        // --------------------------------------------------------------------
        // Create pipe: Burst -> Li
        // --------------------------------------------------------------------
        for (nTs = 0; nTs < g_demoEnv.nBurstHwyTimeslots; nTs++) {
          result = OtxDrvCreatePipe(64, &hPipe); DemoCheckRc(result);
          memset(sqBurstTimeSlotMask, 0, sizeof(sqBurstTimeSlotMask));
          sqBurstTimeSlotMask[nTs] = 0xFF;

          // -------------------------
          // Connect pipe: Burst -> Li
          // -------------------------
          result = OtxDrvConnectPipe(hPipe, OTX_PIPE_INPUT,  g_hPhysDevs.hTss, sqTssBurstHighway[iHighwayNo], sqBurstTimeSlotMask, g_demoEnv.nBurstHwyTimeslots);
          DemoCheckRc(result);
          
          result = OtxDrvConnectPipe(hPipe, OTX_PIPE_OUTPUT, g_hPhysDevs.hTss, sqTssLiHighway[iHighwayNo], sqBurstTimeSlotMask, g_demoEnv.nBurstHwyTimeslots);
          DemoCheckRc(result);


          // --------------------------------------------------------------------
          // Create pipe: Li -> Burst
          // --------------------------------------------------------------------
          result = OtxDrvCreatePipe(64, &hPipe); DemoCheckRc(result);
          
          // -------------------------
          // Connect pipe: Li -> Burst
          // -------------------------
          result = OtxDrvConnectPipe(hPipe, OTX_PIPE_INPUT,  g_hPhysDevs.hTss, sqTssLiHighway[iHighwayNo], sqBurstTimeSlotMask, g_demoEnv.nBurstHwyTimeslots);
          DemoCheckRc(result);
          
          result = OtxDrvConnectPipe(hPipe, OTX_PIPE_OUTPUT, g_hPhysDevs.hTss, sqTssBurstHighway[iHighwayNo], sqBurstTimeSlotMask, g_demoEnv.nBurstHwyTimeslots);
          DemoCheckRc(result);
        }
      }

      // Open E1T1 devices
      switch (g_demoEnv.nBoardType) {
      case OTX_DEVICE_THOR_2_PCI_PLUS:
      case OTX_DEVICE_THOR_2_PCI_EXPRESS:
      case OTX_DEVICE_THOR_1_PCI_PLUS:
        for (nLi = 0; nLi < coLi; nLi++) {
          result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController, OTX_DEVICE_LI_T1E1, nLi, nLi, g_hEventQueue, &hLi[nLi]); DemoCheckRc(result);
        }
        break;
      case OTX_DEVICE_THOR_4_PCI_PLUS:
      case OTX_DEVICE_THOR_8_PCI_PLUS:
        for (nQuadE1T1 = 0; nQuadE1T1 < coQuadE1T1; nQuadE1T1++) {
          result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController, OTX_DEVICE_QUAD_T1E1, nQuadE1T1, nQuadE1T1, g_hEventQueue, &hQuadE1T1[nQuadE1T1]); DemoCheckRc(result);
          result = OtxDrvEnable(hQuadE1T1[nQuadE1T1], OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
        }
        for (nLi = 0; nLi < coLi; nLi++) {
          result = OtxDrvOpenPhysicalDevice(hQuadE1T1[nLi/4], OTX_DEVICE_LI_T1E1, nLi%4, nLi, g_hEventQueue, &hLi[nLi]); DemoCheckRc(result);
        }
        break;
      default:
        DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
        OsaConsoleExit();
        exit(1);
        break;
      }
        
      // Initialize E1 Configuration Data
      sE1CfgData.eTxLineCode                = OTX_E1_LC_HDB3;
      sE1CfgData.eRxLineCode                = OTX_E1_LC_HDB3;
      sE1CfgData.eTxFrameFormat             = OTX_E1_FF_CRC4_MULTIFRAME;
      sE1CfgData.eRxFrameFormat             = OTX_E1_FF_CRC4_MULTIFRAME;
      sE1CfgData.eAisDetection              = OTX_E1_AIS_ETS300233;
      sE1CfgData.bfSiBits                   = 0x3;
      sE1CfgData.bfSaBits                   = 0x1F;
      sE1CfgData.bExtendedHDB3errorDetect   = OTX_FALSE;
      sE1CfgData.bAutoRegainMultiframe      = OTX_FALSE;
      sE1CfgData.bAutoResync                = OTX_TRUE;
      sE1CfgData.bAutoCrcBitInversion       = OTX_FALSE;
      sE1CfgData.bAutoRemoteAlarm           = OTX_FALSE;
      sE1CfgData.nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775
      sE1CfgData.nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775
      sE1CfgData.nLineLength                = 100;
      sE1CfgData.bHighTxPower               = OTX_FALSE;
      sE1CfgData.bReceiveEqualizer          = OTX_FALSE;;
      
      // Configure Transceivers
      for (nLi = 0; nLi < coLi; nLi++) {
        OTX_UINT8 sqTimeSlotMask[32];
        // Set termination for the T1/E1 Front End
        result = OtxT1E1SetLineTermination(hLi[nLi], OTX_T1E1_TERM_100_OHM); DemoCheckRc(result);
        // E1 mode
        result = OtxE1SetConfiguration(hLi[nLi], &sE1CfgData);  DemoCheckRc(result);
        result = OtxT1E1SetLiMode(hLi[nLi], OTX_T1E1_MODE_E1); DemoCheckRc(result);
        
        if (g_demoEnv.bUseExternalLoop) {
          // Use external loop cables
          result = OtxT1E1SetLineLoop(hLi[nLi], OTX_T1E1_LOOPS_OFF); DemoCheckRc(result);
        } else {
          // Tell the E1T1 device to loop its internal highway back to the internal highway
          result = OtxT1E1SetLineLoop(hLi[nLi], OTX_T1E1_LOCAL_LOOP); DemoCheckRc(result);
        }
        
        // Although the timeslots connections are hardcoded on this card, we still need to connect
        // to the timeslots of the Li devices that we want to control. If we don't connect any pipes 
        // to the Li devices then it will send an Idle code on each timeslot.
        
        memset(sqTimeSlotMask, 0xFF, sizeof(sqTimeSlotMask));
        
        result = OtxDrvCreatePipe(64*32, &hPipe); DemoCheckRc(result);
        result = OtxDrvConnectPipe(hPipe, OTX_PIPE_OUTPUT, hLi[nLi],  -1, sqTimeSlotMask, 32); DemoCheckRc(result);
        
        result = OtxDrvEnable(hLi[nLi], OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
      }

    } else if (g_demoEnv.bUseDspLoop) {
      Demo2IoInLoopAllTsS sIn;
      OTX_UINT32 nBytesReturned;
      OTX_UINT32 nTsPerDspHwy;
      OTX_UINT32 nDspHwyTsInUse;
      OTX_UINT32 nTotalDspHwyTs;
      OTX_UINT16 nDemo2Rc;
      char *pszDspProgram = "Dsp55Demo2.out";
      OTX_UINT32 nDsp;
      OTX_UINT32 nDspBlock;
      OTX_HANDLE hAsmDsp[4];
      OTX_HANDLE hUserApp[4];
      OTX_INT32 iBurstHwy;
      OTX_UINT8 sqBurstTimeSlotMask[OTX_PIPE_TIMESLOTS_MAX];
      OTX_UINT8 sqDspTimeSlotMask[OTX_PIPE_TIMESLOTS_MAX];
      OTX_UINT16 nDspCommand = 0;
      OTX_UINT32 nBurstTs = 0xFFFFFFFF;
      
      // Get a handle to to the ASM Daughter Board
      // Try to open a Vidar-5x4-ASM-EX board
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController, OTX_DEVICE_VIDAR_5X4_ASM_EX, 0, 0, g_hEventQueue, &g_hAsm); 
      if (OTX_SUCCESS(result)) {
        pszDspProgram = "DspDemo2c5410a.out";
        nDspCommand = 0x010A;   // #define DEMO2_IOCTL_LOOP_ALL_TS for Dsp54x
        goto ASM_IS_OPENED;
      }
      
      // Try to open a Vidar-5x4-ASM-PRO board
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController, OTX_DEVICE_VIDAR_5X4_ASM_PRO, 0, 0, g_hEventQueue, &g_hAsm); 
      if (OTX_SUCCESS(result)) {
        pszDspProgram = "DspDemo2c5416.out";
        nDspCommand = 0x010A;   // #define DEMO2_IOCTL_LOOP_ALL_TS for Dsp54x
        goto ASM_IS_OPENED;
      }
      
      // Try to open a Vidar-55x4-ASM board
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController, OTX_DEVICE_VIDAR_55x4_ASM, 0, 0, g_hEventQueue, &g_hAsm); 
      if (OTX_SUCCESS(result)) {
        pszDspProgram = "Dsp55Demo2.out";
        nDspCommand = DEMO2_IOCTL_LOOP_ALL_TS; // Dsp55x
        goto ASM_IS_OPENED;
      }
      
      DemoLogf("Error: Could not open a supported ASM board for this application.\n");
      DemoExitQuestion();
      
ASM_IS_OPENED:
      
      for (nDsp = 0; nDsp < 4; nDsp++) {
        DemoLogf("Open DSP%d\n", nDsp);
        result = OtxDrvOpenPhysicalDevice(g_hAsm, OTX_DEVICE_DSP, nDsp, nDsp, g_hEventQueue, &hAsmDsp[nDsp]); DemoCheckRc(result);
        
        DemoLogf("Loading DSP Program: %s\n", pszDspProgram);
        result = OtxDspRunProgram(hAsmDsp[nDsp], pszDspProgram, "_c_int00"); DemoCheckRc(result);
        
        DemoLogf("Creating UserApp\n");
        result = OtxDrvCreateLogicalDevice(hAsmDsp[nDsp], OTX_LDEVICE_USER_APPLICATION, g_hEventQueue, nDsp, &hUserApp[nDsp]); DemoCheckRc(result);
        
        // Instruct the DSP program to loop all timeslots
        sIn.bLoopAllTs = 1;
        
        DemoLogf("Sending DEMO2_IOCTL_LOOP_ALL_TS command\n");
        result = OtxDspIoControl(
          hUserApp[nDsp], 
          nDspCommand, 
          &sIn, sizeof(sIn), 
          &nDemo2Rc, sizeof(nDemo2Rc), 
          &nBytesReturned, 
          OTX_TASK_SYNC, 
          5000 //OTX_TIME_INFINITY
          ); 
        DemoCheckRc(result);
        
      }
      
      // Get DSP Highway Speed
      OtxDrvGetAttributeValueUint32(hAsmDsp[0], OTX_ATTR_DSP_HIGHWAY_SPEED, &g_nDspHighwaySpeed); 
      DemoLogf("DSP[0] highway speed is %d kbps\n", g_nDspHighwaySpeed);
      nTsPerDspHwy = g_nDspHighwaySpeed / 64;
      nTotalDspHwyTs = nTsPerDspHwy * 4;
      if ((OTX_INT32)nTotalDspHwyTs < g_demoEnv.coBurstHighways * g_demoEnv.nBurstHwyTimeslots) {
        nDspHwyTsInUse = ((g_demoEnv.coBurstHighways * g_demoEnv.nBurstHwyTimeslots) - nTotalDspHwyTs) / 4;
      } else {
        nDspHwyTsInUse = nTsPerDspHwy;
      }
      
      // Create the Pipes (one for each direction) between the Burst Device and the 4 DSPs
      DemoLogf("Creating pipes for Burst Sender/Receiver [coBurstHighways=%d] [nBurstHwyTimeslots=%d] [nDspHwyTsInUse=%d]\n", g_demoEnv.coBurstHighways, g_demoEnv.nBurstHwyTimeslots, nDspHwyTsInUse);
      for (nDsp = 0, nDspBlock = 0, iBurstHwy = 0; (iBurstHwy < g_demoEnv.coBurstHighways) && (nDsp < 4); iBurstHwy++, nDspBlock = 0) {
        OTX_UINT32 nDspTs;
        
        nBurstTs = 0;
        while ((OTX_INT32)nBurstTs < g_demoEnv.nBurstHwyTimeslots) {
          // --------------------------------------------------------------------
          // Connect all timeslots that we will use from this DSP
          // --------------------------------------------------------------------
          DemoLogf("Connecting Dsp%d [nDspBlock=%d]\n", nDsp, nDspBlock);

          for (nDspTs = 0; nDspTs < nDspHwyTsInUse; nDspTs++) {
            OTX_HANDLE hPipe;
            
            memset(sqBurstTimeSlotMask, 0, sizeof(sqBurstTimeSlotMask));
            nBurstTs = nDspBlock * nDspHwyTsInUse + nDspTs;
            sqBurstTimeSlotMask[nBurstTs] = 0xFF;
            
            memset(sqDspTimeSlotMask, 0, sizeof(sqDspTimeSlotMask));
            sqDspTimeSlotMask[nDspTs] = 0xFF;
            
            DemoLogf("[bHwy=%d, nDsp=%d, nDspBlock=%d, nBurstTs=%d, nDspTs=%d] \n", iBurstHwy, nDsp, nDspBlock, nBurstTs, nDspTs);
            // --------------------------------------
            // Create and Connect pipe: Burst -> Dsp
            // --------------------------------------
            result = OtxDrvCreatePipe(64, &hPipe); DemoCheckRc(result);
            
            result = OtxDrvConnectPipe(hPipe, OTX_PIPE_INPUT,  g_hPhysDevs.hBurst, iBurstHwy, sqBurstTimeSlotMask, g_demoEnv.nBurstHwyTimeslots);
            DemoCheckRc(result);
            
            result = OtxDrvConnectPipe(hPipe, OTX_PIPE_OUTPUT, hAsmDsp[nDsp], -1, sqDspTimeSlotMask, nTsPerDspHwy);
            DemoCheckRc(result);
            
            // --------------------------------------------------------------------
            // Create and Connect pipe: Dsp -> Burst
            // --------------------------------------------------------------------
            result = OtxDrvCreatePipe(64, &hPipe); DemoCheckRc(result);
            
            result = OtxDrvConnectPipe(hPipe, OTX_PIPE_INPUT,  hAsmDsp[nDsp], -1, sqDspTimeSlotMask, nTsPerDspHwy);
            DemoCheckRc(result);
            
            result = OtxDrvConnectPipe(hPipe, OTX_PIPE_OUTPUT, g_hPhysDevs.hBurst, iBurstHwy, sqBurstTimeSlotMask, g_demoEnv.nBurstHwyTimeslots);
            DemoCheckRc(result);
            
          }
          
          nDsp++;
          nDspBlock++;
          nDspTs = 0;
          nBurstTs = nDspBlock * nDspHwyTsInUse + nDspTs;
        }
        
      }

      assert((int)nBurstTs == (int)g_demoEnv.nBurstHwyTimeslots); // make sure we filled up the entire burst highway

      // In case we have more burst timeslots than we have DSP timeslots we will loop the rest of the burst highways to themselves.
      for (; (iBurstHwy < g_demoEnv.coBurstHighways); iBurstHwy++) {
        
        for (nBurstTs = 0; nBurstTs < (OTX_UINT32)g_demoEnv.nBurstHwyTimeslots; nBurstTs++) {
          OTX_HANDLE hPipe;
          
          memset(sqBurstTimeSlotMask, 0, sizeof(sqBurstTimeSlotMask));
          sqBurstTimeSlotMask[nBurstTs] = 0xFF;
          
          DemoLogf("[bHwy=%d, nBurstTs=%d looped back to itself] \n", iBurstHwy, nBurstTs);
          // --------------------------------------
          // Create and Connect pipe: Burst -> Burst
          // --------------------------------------
          result = OtxDrvCreatePipe(64, &hPipe); DemoCheckRc(result);
          
          result = OtxDrvConnectPipe(hPipe, OTX_PIPE_INPUT,  g_hPhysDevs.hBurst, iBurstHwy, sqBurstTimeSlotMask, g_demoEnv.nBurstHwyTimeslots);
          DemoCheckRc(result);
          
          result = OtxDrvConnectPipe(hPipe, OTX_PIPE_OUTPUT,  g_hPhysDevs.hBurst, iBurstHwy, sqBurstTimeSlotMask, g_demoEnv.nBurstHwyTimeslots);
          DemoCheckRc(result);
          
        }
      }

    }else if (g_demoEnv.bUseH100Loop) {

      //DemoConnectH100();
      DemoLogf("H.100 mode enabled connect pipe when all cards are active\n");

    }
    else {
      // Loop the burst highway via the TSS device (since these boards all have a TSS device).
      OTX_INT32 iHighwayNo;
      OTX_HANDLE hPipe;
      OTX_UINT8 sqBurstTimeSlotMask[OTX_PIPE_TIMESLOTS_MAX];
      
      
      // Create the Pipes (one for each direction) between the Burst Device and the LIs
      for (iHighwayNo = 0; iHighwayNo < g_demoEnv.coBurstHighways; iHighwayNo++) {
        
        memset(sqBurstTimeSlotMask, 0, sizeof(sqBurstTimeSlotMask));
        memset(sqBurstTimeSlotMask, 0xFF, g_demoEnv.nBurstHwyTimeslots);
        
        DemoLogf("Creating pipe for Burst Sender/Receiver [highway=%d]\n", iHighwayNo);
        
        // --------------------------------------------------------------------
        // Create an incoming and outgoing pipe 
        // --------------------------------------------------------------------
        result = OtxDrvCreatePipe(64*g_demoEnv.nBurstHwyTimeslots, &hPipe); DemoCheckRc(result);
        
        // -------------------------
        // Connect pipe: Burst -> Burst
        // -------------------------
        result = OtxDrvConnectPipe(hPipe, OTX_PIPE_INPUT,  g_hPhysDevs.hBurst, iHighwayNo, sqBurstTimeSlotMask, g_demoEnv.nBurstHwyTimeslots);
        DemoCheckRc(result);
        
        result = OtxDrvConnectPipe(hPipe, OTX_PIPE_OUTPUT, g_hPhysDevs.hBurst, iHighwayNo, sqBurstTimeSlotMask, g_demoEnv.nBurstHwyTimeslots);
        DemoCheckRc(result);
        
      }
    }
    break;

  case OTX_DEVICE_HUGIN_2_PC104:
      DemoLogf("---OTX_DEVICE_HUGIN_2_PC104\n");
    
  case OTX_DEVICE_THOR_2_EXPRESSCARD:
  case OTX_DEVICE_THOR_4_EXPRESSCARD:
    {
      int i;

      if (g_demoEnv.bUseDspLoop) {
        OTX_HANDLE hDualE1T1;
        Demo2IoInLoopAllTsS sIn;
        OTX_UINT32 nBytesReturned;
        OTX_UINT16 nDemo2Rc;
        char *pszDspProgram       = "Dsp55Demo2NoExtMem.out";
        char *pszDspProgramExtMem = "Dsp55Demo2.out";

#if 1 // This should not be needed. A bug in the 2B firmware?
        if(g_demoEnv.nBoardType == OTX_DEVICE_THOR_2_EXPRESSCARD) {
          result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController, OTX_DEVICE_DUAL_T1E1, 0, 0, g_hEventQueue, &hDualE1T1); DemoCheckRc(result);
          result = OtxDrvEnable(hDualE1T1, OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
        }
        else {
          if(g_demoEnv.nBoardType == OTX_DEVICE_HUGIN_2_PC104) {
            // Open the Line interfaces
            for (i = 0; i < 2; i++) {
              DemoLogf("---OTX_DEVICE_HUGIN_2_PC104 OTX_DEVICE_LI_T1E1 %d\n", i);
              result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController, OTX_DEVICE_LI_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(result);
              result = OtxDrvEnable(g_hPhysDevs.hLi[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
            }
          }
          else {
            result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController, OTX_DEVICE_QUAD_T1E1, 0, 0, g_hEventQueue, &g_hPhysDevs.hLi[0]); DemoCheckRc(result);
            result = OtxDrvEnable(g_hPhysDevs.hLi[0], OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
          }

          DemoLogf("Configure Burst to DSP data path\n");
          // we'll source the Falc from the Burst device, and source the burst device from the FALC
          //  Falc <-- Burst Send (P2S)
          result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hThorController, OTX_ATTR_BRD_FALC_SOURCED, OTX_BRD_FALC_SOURCED_P2S); DemoCheckRc(result);
          //  Burst Recv (S2P) <-- DSP
          result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hThorController, OTX_ATTR_BRD_S2P_SOURCED, OTX_BRD_S2P_SOURCED_DSP);  DemoCheckRc(result);
          //  DSP <-- Burst Recv (S2P)
          result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hThorController, OTX_ATTR_BRD_DSP_SOURCED, OTX_BRD_DSP_SOURCED_P2S);  DemoCheckRc(result);
        }
#endif
        
        for(i=0; i<g_demoEnv.nDspNo; i++) {
          DemoLogf("Open DSP[%d]\n", i);
          result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController, OTX_DEVICE_DSP, i, i, g_hEventQueue, &g_hDsp[i]); DemoCheckRc(result);

          // Get DSP Highway Speed
          OtxDrvGetAttributeValueUint32(g_hDsp[i], OTX_ATTR_DSP_HIGHWAY_SPEED, &g_nDspHighwaySpeed); 
          DemoLogf("DSP highway speed is %d kbps\n", g_nDspHighwaySpeed);
  #if 0 // temporary
          DemoLogf("Disable g_hPhysDevs.hBurst\n");
          result = OtxDrvDisable(g_hPhysDevs.hBurst, OTX_TASK_SYNC, 5000L);
  #endif

          if(g_demoEnv.nBoardType == OTX_DEVICE_HUGIN_2_PC104) {
            DemoLogf("Loading DSP Program: %s\n", pszDspProgramExtMem);

            result = OtxDspRunProgram(g_hDsp[i], pszDspProgramExtMem, "_c_int00"); 
            DemoCheckRc(result);

          }
          else {
            DemoLogf("Loading DSP Program: %s\n", pszDspProgram);
            result = OtxDspRunProgram(g_hDsp[i], pszDspProgram, "_c_int00"); DemoCheckRc(result);
          }

          result = OtxDrvCreateLogicalDevice(g_hDsp[i], OTX_LDEVICE_USER_APPLICATION, g_hEventQueue, 0, &g_hUserApp[i]); DemoCheckRc(result);

          // Instruct the DSP program (Dsp55Demo2NoExtMem.out) to loop all timeslots
          sIn.bLoopAllTs = 1;
        
          result = OtxDspIoControl(
            g_hUserApp[i], 
            DEMO2_IOCTL_LOOP_ALL_TS, 
            &sIn, sizeof(sIn), 
            &nDemo2Rc, sizeof(nDemo2Rc), 
            &nBytesReturned, 
            OTX_TASK_SYNC, 
            5000 //OTX_TIME_INFINITY
            ); 
          DemoCheckRc(result);

        } // for g_demoEnv.nDspNo        
      } else {
        // Loop the burst data in the T1E1 device device (62 of the 128 timeslots)
        OTX_HANDLE hDualE1T1;
        OTX_HANDLE hLi[2];
        OTX_INT32 i;
        OtxE1ConfigOptionsS sE1CfgData;
        OTX_UINT8 sqTimeSlotMask[32];
        OTX_HANDLE hPipe;

        // We'll source the Falc from the Burst device, 
        //  Falc <-- Burst Send (P2S)
        result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hThorController, OTX_ATTR_BRD_FALC_SOURCED, OTX_BRD_FALC_SOURCED_P2S); DemoCheckRc(result);

        // Open E1T1 devices
        if(g_demoEnv.nBoardType == OTX_DEVICE_THOR_2_EXPRESSCARD) {
          result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController, OTX_DEVICE_DUAL_T1E1, 0, 0, g_hEventQueue, &hDualE1T1); DemoCheckRc(result);
          result = OtxDrvEnable(hDualE1T1, OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
        }
        else {
          result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController, OTX_DEVICE_QUAD_T1E1, 0, 0, g_hEventQueue, &hDualE1T1); DemoCheckRc(result);
          result = OtxDrvEnable(hDualE1T1, OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
        }

        for (i = 0; i < 2; i++) {
          result = OtxDrvOpenPhysicalDevice(hDualE1T1, OTX_DEVICE_LI_T1E1, i, i, g_hEventQueue, &hLi[i]); DemoCheckRc(result);
        }
        
        // Initialize E1 Configuration Data
        sE1CfgData.eTxLineCode                = OTX_E1_LC_HDB3;
        sE1CfgData.eRxLineCode                = OTX_E1_LC_HDB3;
        sE1CfgData.eTxFrameFormat             = OTX_E1_FF_CRC4_MULTIFRAME;
        sE1CfgData.eRxFrameFormat             = OTX_E1_FF_CRC4_MULTIFRAME;
        sE1CfgData.eAisDetection              = OTX_E1_AIS_ETS300233;
        sE1CfgData.bfSiBits                   = 0x3;
        sE1CfgData.bfSaBits                   = 0x1F;
        sE1CfgData.bExtendedHDB3errorDetect   = OTX_FALSE;
        sE1CfgData.bAutoRegainMultiframe      = OTX_FALSE;
        sE1CfgData.bAutoResync                = OTX_TRUE;
        sE1CfgData.bAutoCrcBitInversion       = OTX_FALSE;
        sE1CfgData.bAutoRemoteAlarm           = OTX_FALSE;
        sE1CfgData.nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775
        sE1CfgData.nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775
        sE1CfgData.nLineLength                = 100;
        sE1CfgData.bHighTxPower               = OTX_FALSE;
        sE1CfgData.bReceiveEqualizer          = OTX_FALSE;;
        
        // Configure Transceivers
        for (i = 0; i < 2; i++) {
          // Set termination for the T1/E1 Front End
          result = OtxT1E1SetLineTermination(hLi[i], OTX_T1E1_TERM_100_OHM); DemoCheckRc(result);
          // E1 mode
          result = OtxE1SetConfiguration(hLi[i], &sE1CfgData);  DemoCheckRc(result);
          result = OtxT1E1SetLiMode(hLi[i], OTX_T1E1_MODE_E1); DemoCheckRc(result);
          
          if (g_demoEnv.bUseExternalLoop) {
            // Use external loop cables
            result = OtxT1E1SetLineLoop(hLi[i], OTX_T1E1_LOOPS_OFF); DemoCheckRc(result);
          } else {
            // Tell the E1T1 device to loop its internal highway back to the internal highway
            result = OtxT1E1SetLineLoop(hLi[i], OTX_T1E1_LOCAL_LOOP); DemoCheckRc(result);
          }
          
          // Although the timeslots connections are hardcoded on this card, we still need to connect
          // to the timeslots of the Li devices that we want to control. If we don't connect any pipes 
          // to the Li devices then it will send an Idle code on each timeslot.
          
          memset(sqTimeSlotMask, 0xFF, sizeof(sqTimeSlotMask));
          
          result = OtxDrvCreatePipe(64*32, &hPipe); DemoCheckRc(result);
          result = OtxDrvConnectPipe(hPipe, OTX_PIPE_OUTPUT, hLi[i],  -1, sqTimeSlotMask, 32); DemoCheckRc(result);
          
          result = OtxDrvEnable(hLi[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
        }

      }
      
    }
    break;

  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    OsaConsoleExit();
    exit(1);
    break;
  }

}


/*******************************************************************************

  @func

  Displays a message on the screen 

  @comm

*******************************************************************************/
static void DemoPrintf(char *formatStr, ...)
{
  va_list argList;

  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);

  va_end(argList);
}

static void DemoCheckRc(IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  if (!g_exitFlag) {
    if ( OTX_FAILURE(result) ) {
      OTX_RESULT otxRc;
      char msgStr[256];
      
      // Convert the result code to a more descriptive string
      otxRc = OtxDrvResultCode2String(OTX_DRIVER_DEVICE_HANDLE, result, 256, &msgStr[0]);
      if (OTX_FAILURE(otxRc)) {
        strcpy(msgStr, "Cannot translate this result code.");
      }
      
      DemoLogf("ERROR(%#lX [%d.%d]): %s \n", OTX_RESULT_CODE(result), OTX_RESULT_SOURCE_ID(result), OTX_RESULT_LINE_NUM(result), msgStr);
      
      DemoExitQuestion();
    }
  }
}


/*******************************************************************************
  @func
    Prompts whether to exit the program. Usually called after an error
    has been reported.

*******************************************************************************/
void DemoExitQuestion()
{
  short c;

  if (g_demoEnv.bExitOnError) {
    c = 'y';
  } else {
    DemoLogf("Do you want to exit (Y/N) ");
    c = OsaConsoleGetCh();
  }
  if (c == 'y' || c == 'Y') {
    // Clean up
    OtxDrvReleaseLib();
    OsaConsoleExit();
    if (g_nOverRunErrors || g_nUnderRunErrors) {
      exit(2);
    } else {
      exit(1);
    }
  }
}


/*******************************************************************************
  @func
    Prints a timestamp in a MM-DD hh:mm:ss:mmm format, where

    MM  = month
    DD  = day (of month)
    hh  = hour of day
    mm  = minutes of day
    ss  = seconds of day
    mmm = milliseconds

*******************************************************************************/
void DemoPrintTimeStamp(
  OTX_DATETIME nTimeStamp
)
{
  OTX_DATETIME nTimeStampMs;
  OTX_DATETIME nTimeStampSec;
  struct Osa_tm *ptCurrentDayAndTime;
  Osa_time_t nCurrentTimeSecs;

  // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
  #if defined(__linux__)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000LL;  
  #elif defined(WIN32)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;
  #endif

  // Convert to seconds since Jan 1, 1970 (so we can use this time with OsaStdLocalTime())
  nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;

  // Calculate number of milliseconds (since number of seconds is not a very accurate way to report timestamp)
  nTimeStampMs.nUint64 = nTimeStamp.nUint64 / 10000 - (nTimeStampSec.nUint64 * 1000);

  nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
  ptCurrentDayAndTime = OsaStdLocalTime(&nCurrentTimeSecs);

  if (ptCurrentDayAndTime) {

    DemoLogf("%.2d-%.2d %.2d:%.2d:%.2d.%.3d ",
      ptCurrentDayAndTime->tm_mon + 1,
      ptCurrentDayAndTime->tm_mday,
      ptCurrentDayAndTime->tm_hour,
      ptCurrentDayAndTime->tm_min,
      ptCurrentDayAndTime->tm_sec,
      nTimeStampMs.sUint32.nLow
      );
  } else {
    DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
  }

}


/*******************************************************************************
  @func
    Idle Function that check for Events from the OTX Driver
*******************************************************************************/
void *DemoIdleFunction(
  void *lpParameter   // thread data
)
{
  OtxEventDataS eventData;
  OTX_RESULT result;

  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);

    do {
      result = OtxDrvGetEventData(g_hEventQueue, &eventData); DemoCheckRc(result);
      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event

        switch (eventData.m_eDeviceType) {
          case OTX_DEVICE_BURST:
            switch (eventData.m_nCode) {
            case OTX_BURST_RECEIVER_EC_DATA_AVAILABLE:
              DemoBurstReceiver(eventData.m_nRequestId);
              continue;
            case OTX_BURST_SENDER_EC_BUF_DEPLETED:
              DemoBurstSender(eventData.m_nRequestId);
              continue;
            case OTX_BURST_EC_BUFFER_OVERRUN:  // Receiver data overrun error - data lost
              g_nOverRunErrors++;
              DemoPrintEvent(&eventData);
              break;
            case OTX_BURST_EC_BUFFER_UNDERRUN: // Sender data underrun error
              g_nUnderRunErrors++;
              DemoPrintEvent(&eventData);
              break;
            default:
              break;
            }
            break;
          default:
            break;
        }

        // Printf the event data.
        // In a real application branch based on the data
        DemoPrintEvent(&eventData);
      }
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);

  }

  return 0;
}


/*******************************************************************************
  @func
    Burst Sender Function
*******************************************************************************/
void DemoBurstSender(OTX_UINT32 nSourceId)
{
  OTX_RESULT  otxRc;
  OTX_UINT32  ui32;


  // validate parameters
  if (nSourceId < 0 || nSourceId > (OTX_UINT32)g_demoEnv.coBurstHighways) {
    DemoLogf("DemoBurstSender: Invalid nSourceId(%d)!\n", nSourceId);
    return;
  }

  if (!g_hLogicDevs.mBurstSender[nSourceId].bActive) {
    // we're not active, so just return
    return;
  }

START:
  if (g_hLogicDevs.mBurstSender[nSourceId].pTxFile) {
    // Prime the transmit buffer with 0xFF (only really needed for single TS transmission)
    memset(&g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf[0], 0xFF, sizeof(g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf) / sizeof(g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf[0]));

    if (g_hLogicDevs.mBurstSender[nSourceId].nTsToSend == -1) {
      // All TS
      g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf = fread(
        &g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf,
        sizeof(g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf[0]),
        sizeof(g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf) / sizeof(g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf[0]),
        g_hLogicDevs.mBurstSender[nSourceId].pTxFile
        );
      
    } else {
      // Single TS
      OTX_UINT8 singleTsBuf[DEMO_BURST_BUF_SIZE / 32]; // On some boards this is bigger than we actually use
      int cbSingleTsBuf;
      int byteCount = 0;
      int cbMaxSingleTsBuf = DEMO_BURST_BUF_SIZE / g_demoEnv.nBurstHwyTimeslots;
      int x;
      
      cbSingleTsBuf = fread(
        &singleTsBuf,
        sizeof(singleTsBuf[0]),
        cbMaxSingleTsBuf,
        g_hLogicDevs.mBurstSender[nSourceId].pTxFile
        );

      if (cbSingleTsBuf > 0) {
        if ((cbSingleTsBuf > 0) && (cbSingleTsBuf < cbMaxSingleTsBuf)) {
          // We are at the end of the file, and we don't have a full buffer to send.
          // Pad with 0xFF at the end of the buffer to make sure that we don't underrun the Burst Sender
          DemoLogf("Pad with 0xFF [%d] at the end of the buffer to make sure that we don't underrun the Burst Sender\n", (cbMaxSingleTsBuf - cbSingleTsBuf) / sizeof(singleTsBuf[0]));
          memset(&singleTsBuf[cbSingleTsBuf], 0xFF, (cbMaxSingleTsBuf - cbSingleTsBuf) / sizeof(singleTsBuf[0]) );
          cbSingleTsBuf = cbMaxSingleTsBuf;
        }
        
        // Dispurse the single TS to be transmitted in the whole transmit buffer 
        byteCount = 0;
        //DemoLogf("g_demoEnv.coBurstHighways=%d, g_hLogicDevs.mBurstSender[nSourceId].nTsToSend=%d \n", g_demoEnv.nCurrentHwy, g_demoEnv.coLi, g_demoEnv.coBurstHighways, g_hLogicDevs.mBurstSender[nSourceId].nTsToSend);
        for ( x = g_hLogicDevs.mBurstSender[nSourceId].nTsToSend; 
        (byteCount < cbSingleTsBuf) && (x < DEMO_BURST_BUF_SIZE); 
        x += g_demoEnv.nBurstHwyTimeslots
          ) {
          g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf[x] = singleTsBuf[byteCount++];
        }
        
        assert(byteCount == cbSingleTsBuf);
          
      }

      g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf = cbSingleTsBuf * g_demoEnv.nBurstHwyTimeslots;

    }
    
    if (g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf > 0) {
      ui32 = g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf;

      assert(g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf == DEMO_BURST_BUF_SIZE);

      otxRc = OtxBurstSendData(g_hPhysDevs.hBurst, nSourceId, g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf, &ui32);
      DemoCheckRc(otxRc);
      
      // check the amount buffered w/ how much we requested to be buffered
      //  (since we set the trigger amount to the buffer size, there should
      //    have been room to buffer this, so these two should be the same).
      //assert(ui32 == g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf);
      if (ui32 != g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf) {
        DemoLogf("DemoBurstSender[%d]: ui32 (%d) != g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf (%d)\n", nSourceId, ui32 != g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf);
      }
      
    } else {
      
      if (g_hLogicDevs.mBurstSender[nSourceId].bFileLoop) {
        if (g_hLogicDevs.mBurstSender[nSourceId].bVerbose) {
          DemoLogf("DemoBurstSender[%d]: Transmission complete. Starting file from the beginning.\n", nSourceId);
        }
        rewind(g_hLogicDevs.mBurstSender[nSourceId].pTxFile);
        goto START;
      } else {
        DemoLogf("DemoBurstSender[%d]: Transmission complete. Closing file\n", nSourceId);
        fclose(g_hLogicDevs.mBurstSender[nSourceId].pTxFile);
        g_hLogicDevs.mBurstSender[nSourceId].pTxFile = NULL;
        g_hLogicDevs.mBurstSender[nSourceId].bActive = OTX_FALSE;
      }
    }
    
  } else {
    // Send a gamma pattern, which means that we increase the value by one for each frame (continuously until aborted)
    int x, y;

    for (x = 0; x < DEMO_BURST_BUF_SIZE; x += g_demoEnv.nBurstHwyTimeslots) {
      for (y = 0; y < g_demoEnv.nBurstHwyTimeslots; y++) {
        g_hLogicDevs.mBurstSender[nSourceId].sqGammaData[y]++;
        g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf[x+y] = g_hLogicDevs.mBurstSender[nSourceId].sqGammaData[y];
      }
    }

    ui32 = g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf;
    otxRc = OtxBurstSendData(g_hPhysDevs.hBurst, nSourceId, g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf, &ui32);
    DemoCheckRc(otxRc);

    // check the amount buffered w/ how much we requested to be buffered
    //  (since we set the trigger amount to the buffer size, there should
    //    have been room to buffer this, so these two should be the same).
    //assert(ui32 == g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf);
    if(ui32 != g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf) {
      DemoLogf("DemoBurstSender[%d]: OtxBurstSendData did not copy entire buffer.\n", nSourceId);
      DemoLogf("   Buffer size sent %d\n", g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf);
      DemoLogf("   Buffer size returned %d\n", ui32);
    }
  }

  return;
}

/*******************************************************************************
  @func
    Burst Receiver Function
*******************************************************************************/
void DemoBurstReceiver(OTX_UINT32 nSourceId)
{
  OTX_RESULT    otxRc;
  OTX_UINT8     buf[DEMO_BURST_BUF_SIZE];
  OTX_UINT32    ui32;
  OTX_UINT64    nFrameRefId;
  OTX_DATETIME  dtFrameRefTime;

  // validate parameters
  if (nSourceId < 0 || nSourceId > 2) {
    DemoLogf("DemoBurstReceiver: Invalid nSourceId(%d)...\n", nSourceId);
    return;
  }

  if (!g_hLogicDevs.mBurstReceiver[nSourceId].bActive) {
    // we're not active, so just return
    return;
  }

  g_hLogicDevs.mBurstReceiver[nSourceId].nDataBlockCounter++;

  // we want the frame reference id and time
  ui32 = sizeof(buf);
  otxRc = OtxBurstReadDataEx(g_hPhysDevs.hBurst, nSourceId, buf, &ui32, &nFrameRefId, &dtFrameRefTime);
  DemoCheckRc(otxRc);

  // check the amount buffered w/ how much we requested to be buffered
  //  (since we set the trigger amount to the buffer size, there should
  //    have been room to buffer this, so these two should be the same).
  assert(ui32 == sizeof(buf));

  // There was a data receive event, add one to the counter
  g_hLogicDevs.mBurstReceiver[nSourceId].nRecvCnt++;
  if(g_hLogicDevs.mBurstReceiver[nSourceId].nRecvCnt == 0) {
    // If count is zero set it to 1. Zero is used to indicate no data received
    g_hLogicDevs.mBurstReceiver[nSourceId].nRecvCnt++;
  }

  // Echo if we're in verbose mode
  if (g_hLogicDevs.mBurstReceiver[nSourceId].bVerbose) {
    int x;

    DemoPrintTimeStamp(dtFrameRefTime);

    // Note: we only display the least significant 32 bits of the frame reference id.
    DemoLogf("Burst Receiver Hwy[%d][%d] (len=%.5d) ID:%08x \n", 
              nSourceId, 
              g_demoEnv.nCurrentHwy, 
              sizeof(buf), 
              (OTX_UINT32) (nFrameRefId & 0xffffffff));
    x = 0; 
    while((x < g_demoEnv.nBurstHwyTimeslots) && (x < DEMO_BURST_BUF_SIZE)) {
      DemoLogf("%.2X ", buf[x]);

      x++;
      if(!(x%16)) {
        DemoLogf("\n"); 
      }
    }

    DemoLogf("\n");
  }

  // Check the received value if we are in Gamma Test Mode
  if (g_hLogicDevs.mBurstReceiver[nSourceId].bGammaTestMode) {
    int x;
    int y;
    
    for (x = 0; x < sizeof(buf); x += g_demoEnv.nBurstHwyTimeslots) {
      for (y = 0; y < g_demoEnv.nBurstHwyTimeslots; y++) {
        if (buf[x+y] == g_hLogicDevs.mBurstReceiver[nSourceId].sqGammaTestData[y]) {
          // Data matched. No action needed.
        } else {
          // Data mismatch. Step error count and resync
          g_hLogicDevs.mBurstReceiver[nSourceId].sqGammaTestData[y] = buf[x+y];
          g_hLogicDevs.mBurstReceiver[nSourceId].nGammaTestErrors[y]++;
        }
        
        // Step the GammaTestData value, which is the next value that we expect
        g_hLogicDevs.mBurstReceiver[nSourceId].sqGammaTestData[y]++;
      }
    }
  }

  
  // write data to a file if desired
  if (g_hLogicDevs.mBurstReceiver[nSourceId].pRxFile && g_hLogicDevs.mBurstReceiver[nSourceId].bActive) {
    int cbWritten;

    if (g_hLogicDevs.mBurstReceiver[nSourceId].nTsToRecord == -1) {
      // All TS
      cbWritten = fwrite(buf, sizeof(buf[0]), sizeof(buf), g_hLogicDevs.mBurstReceiver[nSourceId].pRxFile);
      if (cbWritten != (int) sizeof(buf)) {
        if (!g_exitFlag) {
          DemoLogf("Burst Receiver Hwy[%d] Error: Could not write data to file.  (nFrameLen=%d, cbWritten=%d) \n", nSourceId, sizeof(buf), cbWritten);
        }
      }
    } else {
      // Single TS
      OTX_UINT8 singleTsBuf[DEMO_BURST_BUF_SIZE/32];
      int byteCount = 0;
      int x;

      byteCount = 0;
      for (
          x = g_hLogicDevs.mBurstReceiver[nSourceId].nTsToRecord;
          x < (OTX_INT32) sizeof(buf); 
          x += g_demoEnv.nBurstHwyTimeslots
      ) {
        singleTsBuf[byteCount++] = buf[x];
      }
      cbWritten = fwrite(singleTsBuf, sizeof(singleTsBuf[0]), byteCount, g_hLogicDevs.mBurstReceiver[nSourceId].pRxFile);
      if (cbWritten != byteCount) {
        if (!g_exitFlag) {
          DemoLogf("Burst Receiver Hwy[%d] Error: Could not write data to file.  (byteCount=%d, cbWritten=%d) \n", nSourceId, byteCount, cbWritten);
        }
      } else {
        //DemoLogf("Burst Receiver Hwy[%d] (byteCount=%d, cbWritten=%d) \n", nSourceId, byteCount, cbWritten);
      }
    }
  }
}

void DemoPrintEvent(
/*******************************************************************************
  @func
    Prints a Driver Event.
*******************************************************************************/
  OtxEventDataS *pEventData
)
{
  OTX_RESULT otxRc;
  char msgStr[256];

  switch (pEventData->m_eDeviceType) {
  case OTX_DEVICE_LI_T1E1:
    DemoPrintTimeStamp(pEventData->m_nTimeStamp);

    // Event from a T1E1 device
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      if (pEventData->m_nParam) {
        DemoLogf("E1/T1[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
      } else {
        DemoLogf("E1/T1[%d] Event: %s\n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf("E1/T1[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_DEVICE_BURST:
    DemoPrintTimeStamp(pEventData->m_nTimeStamp);

    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("Burst[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
    } else {
      DemoLogf("Burst[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_LDEVICE_USER_APPLICATION:
    // Events from the Demo2.out logical device
    if (pEventData->m_nCode == OTX_USER_APPLICATION_EC_READ_DATA) {
      OTX_UINT8  buf[8196];
      OTX_UINT32 nBytesReturned;
      OTX_UINT32 nIoControlCode;

      otxRc = OtxDspReadData(pEventData->m_hDevice, OTX_TASK_SYNC, buf, 8196, &nBytesReturned, &nIoControlCode);
      if (otxRc == OTX_S_OK) {
        switch(nIoControlCode) {
        case DEMO2_NOTIFY_DEBUG_DATA:
          {
            Demo2NotifyGetDebugDataS *pDataS = (Demo2NotifyGetDebugDataS *)buf;
            OTX_INT32 i;

            DemoPrintTimeStamp(pEventData->m_nTimeStamp);
            if (nBytesReturned != sizeof(Demo2NotifyGetDebugDataS)) {
              DemoLogf("Unexpected data length for this IoControlCode (nIoControlCode = %d, nBytesReturned=%d)\n", nIoControlCode, nBytesReturned);
            } else {
              DemoLogf("DEMO2_NOTIFY_DEBUG_DATA[%d] (nRequestId=%d, nBytes=%d)\n", pEventData->m_nSourceId, pEventData->m_nRequestId, nBytesReturned);
              DemoLogf("nData0 = %.4X\n", pDataS->nData0);
              DemoLogf("nData1 = %.4X\n", pDataS->nData1);
              DemoLogf("nData2 = %.4X\n", pDataS->nData2);
              DemoLogf("nData3 = %.4X\n", pDataS->nData3);
              DemoLogf("nData4 = %.4X\n", pDataS->nData4);
              DemoLogf("nData5 = %.4X\n", pDataS->nData5);
              DemoLogf("nData6 = %.4X\n", pDataS->nData6);
              DemoLogf("nData7 = %.4X\n", pDataS->nData7);

              for (i = 0; i < 32; i++) {
                if ((i != 0) && (i%8==0)) DemoLogf("\n");
                DemoLogf("%.2d=%.4X ", i, (unsigned short)pDataS->sqBuf[i]);
              }
              DemoLogf("\n");

            }
          }
          break;
        default:
          DemoLogf("Unknown IoControlCode received in OtxDspReadData (nIoControlCode = %d, nBytesReturned=%d, nRequestId=%d)\n", nIoControlCode, nBytesReturned, pEventData->m_nRequestId);
          break;
        }
      } else {
        OTX_RESULT errorCode = otxRc;
        DemoPrintTimeStamp(pEventData->m_nTimeStamp);
        otxRc = OtxDrvResultCode2String(pEventData->m_hDevice, errorCode, 256, &msgStr[0]);
        if (OTX_SUCCESS(otxRc)) {
          DemoLogf("User App[%d] Error: %s \n", pEventData->m_nSourceId, msgStr);
        } else {
          DemoLogf("User App[%d] Invalid Error Code: %ld\n", pEventData->m_nSourceId, errorCode);
        }
      }

    } else {

      DemoPrintTimeStamp(pEventData->m_nTimeStamp);
      otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
      if (OTX_SUCCESS(otxRc)) {
        DemoLogf("User App[%d] Event: %s (nRequestId=%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nRequestId);
      } else {
        DemoLogf("User App[%d] Invalid Event: %ld (nRequestId=%d)\n", pEventData->m_nSourceId, pEventData->m_nCode, pEventData->m_nRequestId);
      }
    }
    break;

  default:
    // Unhandled event
    DemoPrintTimeStamp(pEventData->m_nTimeStamp);
    DemoLogf("Event from an unknown object 0x%X\n", pEventData->m_hDevice);
    break;
  }
}

OTX_INT32 DemoProcessUserCommand(char commandChar)
/*******************************************************************************

  @func

  Processes User Commands

  @comm

*******************************************************************************/
{
  OTX_RESULT result;

  DemoLogf("Key '%c' pressed\n", commandChar);

  switch (commandChar) {

  case '?':
  case 'h':
  case 'H':
    DemoLogf("Cmd  Synopsis\n");
    DemoLogf("===  ========\n");
    DemoLogf(" ?   Display Help text\n");
    DemoLogf(" h   Ditto\n");
    DemoLogf(" q   Quit\n");
    DemoLogf(" 0-1 Select active Burst Highway. Currently %d is selected.\n", g_demoEnv.nCurrentHwy);
    DemoLogf(" a   Abort Burst Hwy[%d] Senders and Receivers\n", g_demoEnv.nCurrentHwy);
    DemoLogf(" r   Start Burst Receiver Hwy[%d] (no file logging)\n", g_demoEnv.nCurrentHwy);
    DemoLogf(" R   Start Burst Receiver Hwy[%d] Prompt for file to recv\n", g_demoEnv.nCurrentHwy);
    DemoLogf(" s   Start Burst Sender Hwy[%d]. Continously send a gamma pattern\n", g_demoEnv.nCurrentHwy);
    DemoLogf(" S   Start Burst Sender Hwy[%d]. Prompt for file to send\n", g_demoEnv.nCurrentHwy);
    DemoLogf(" l   Query Line Status of the two Lis\n");
    DemoLogf(" L   Burst Sender Hwy[%d]. Toggle loop file sends (currently %s)\n", g_demoEnv.nCurrentHwy, g_hLogicDevs.mBurstSender[g_demoEnv.nCurrentHwy].bFileLoop ? "On" : "Off");
    DemoLogf(" e   Enable Burst  (affects all burst highways)\n");
    DemoLogf(" d   Disable Burst (affects all burst highways)\n");
    DemoLogf(" v   Toggle verbose (print received data) setting Burst Receiver[%d]\n", g_demoEnv.nCurrentHwy);
    DemoLogf(" V   Toggle verbose (print send progress) setting Burst Sender[%d]\n", g_demoEnv.nCurrentHwy);
    DemoLogf(" g   Enable/Disable Gamma Test\n");
    DemoLogf(" G   Clear Gamma Test error counters\n");
    DemoLogf(" j   Print Gamma Test Results for current H(condensed form)\n");
    DemoLogf(" J   Print Gamma Test Results (verbose form)\n");
    DemoLogf(" C   Send Constant Byte on TSS Highway\n");
    DemoLogf(" c   Connect H.100 to Burst device\n");
    break;

  case 'g':
    {
      int nActiveBurstReceivers = 0;
      int nHwy = g_demoEnv.nCurrentHwy;
      
      if (g_hLogicDevs.mBurstReceiver[nHwy].bActive == OTX_TRUE) {
        // Burst receiver is running. Ok.

        if (!g_hLogicDevs.mBurstReceiver[nHwy].bGammaTestMode) {
          // Gamma mode is not enabled. So enable it.
          int n;
          memset(g_hLogicDevs.mBurstReceiver[nHwy].sqGammaTestData, 0, g_demoEnv.nBurstHwyTimeslots);
          for (n = 0; n < g_demoEnv.nBurstHwyTimeslots; n++) {
            g_hLogicDevs.mBurstReceiver[nHwy].nGammaTestErrors[n] = 0;
          }
          
          g_hLogicDevs.mBurstReceiver[nHwy].bGammaTestMode = OTX_TRUE;
          DemoLogf("Burst Receiver Hwy[%d] Gamma Test Mode is now enabled\n", nHwy);

        } else {
          // Disable Gamma mode.
          g_hLogicDevs.mBurstReceiver[nHwy].bGammaTestMode = OTX_FALSE;
          DemoLogf("Burst Receiver Hwy[%d] Gamma Test Mode is now disabled\n", nHwy);
        }
      } else {
        DemoLogf("Burst Receiver Hwy[%d] The Burst Receiver is not yet running. Start with 'r' or 'R'.\n");
      }
    }
    break;

  case 'G':  // Zero Gamma Test Error Counters
    {
      int n;
      for (n = 0; n < g_demoEnv.nBurstHwyTimeslots; n++) {
        g_hLogicDevs.mBurstReceiver[g_demoEnv.nCurrentHwy].nGammaTestErrors[n] = 0;
      }          
      // Used to make sure data is received
      g_hLogicDevs.mBurstReceiver[g_demoEnv.nCurrentHwy].nRecvCnt = 0;
      
      DemoLogf("Burst Receiver Hwy[%d] Gamma Test Results are now reset to zero\n", g_demoEnv.nCurrentHwy);
    }
    break;

  case 'j':  // Gamma Test Results (condensed form)
    {
      if (g_hLogicDevs.mBurstReceiver[g_demoEnv.nCurrentHwy].bActive == OTX_TRUE) {
        int y;
        if (g_hLogicDevs.mBurstReceiver[g_demoEnv.nCurrentHwy].bGammaTestMode) {
          OTX_UINT32 nStart;    // Number of dma transfers initiated
          OTX_UINT32 nComplete; // Number of dma transfers completed
          OTX_UINT32 nCheck1;    // Used to check data transfer active
          OTX_UINT32 nCheck2;    // Used to check data transfer active
          OTX_HANDLE hBurst = g_hPhysDevs.hBurst;

          // Need to check to make sure the burst device is receiving data
          OtxDrvGetAttributeValueUint32(hBurst, OTX_ATTR_BURST_STAT_RECV_STARTS, &nStart);
          OtxDrvGetAttributeValueUint32(hBurst, OTX_ATTR_BURST_STAT_RECV_COMPLETES, &nComplete);

          DemoLogf("Burst Receiver Hwy[%d] Gamma Test Results (Condensed Form):\n", g_demoEnv.nCurrentHwy);
          for (y = 0; y < g_demoEnv.nBurstHwyTimeslots; y++) {
            if ((y % 32) == 0) {
              DemoLogf("TS[%.3d] ", y);
            }
            if ((y % 4) == 0) {
              DemoLogf(" "); // Group by 4 timeslots to make it easier to read
            }
            if ((y % 8) == 0) {
              DemoLogf(" "); // Group by 8 timeslots to make it easier to read
            }
            if (g_hLogicDevs.mBurstReceiver[g_demoEnv.nCurrentHwy].nGammaTestErrors[y] > 9) {
              DemoLogf("*"); // Indicates a number larger than 9
            } else {
              DemoLogf("%1d", g_hLogicDevs.mBurstReceiver[g_demoEnv.nCurrentHwy].nGammaTestErrors[y]);
            }
            if (((y+1) % 32) == 0) {
              DemoLogf("\n");
            }
          }
          DemoLogf("\n");

          // wait for a little bit, otherwise, OTX_ATTR_BURST_STAT_RECV_STARTS and OTX_ATTR_BURST_STAT_RECV_COMPLETES don't change yet
          OsaThreadSleep(2);

          // Make sure that the Burst device is receiving data        
          OtxDrvGetAttributeValueUint32(hBurst, OTX_ATTR_BURST_STAT_RECV_STARTS, &nCheck1);
          OtxDrvGetAttributeValueUint32(hBurst, OTX_ATTR_BURST_STAT_RECV_COMPLETES, &nCheck2);
          if((nStart == nCheck1) || (nComplete == nCheck2)) {
            // Count is the same, the burst device is not receiving data
            DemoLogf("\n");
            DemoLogf("Error HW is not receiving data check burst stats\n");
          }
          else {
            DemoLogf("\n");
            DemoLogf("Burst start is receiving data OK\n");
          }

          if(g_hLogicDevs.mBurstReceiver[g_demoEnv.nCurrentHwy].nRecvCnt == 0) {
            // Count is 0, the burst device is not receiving events
            DemoLogf("\n");
            DemoLogf("Error Burst event counter is 0, no data\n");
          }
          else {
            DemoLogf("Burst event counter is OK\n");
          }
        } else {
          DemoLogf("Gamma Test Mode is not enabled\n");
        }
      } else {
        DemoLogf("Burst Receiver Threads are not running. Start them with 'r' or 'R'\n");
      }
    }
    break;

  case 'J':  // Gamma Test Results (verbose form)
    {
      if (g_hLogicDevs.mBurstReceiver[g_demoEnv.nCurrentHwy].bActive == OTX_TRUE) {
        if (g_hLogicDevs.mBurstReceiver[g_demoEnv.nCurrentHwy].bGammaTestMode) {
          int y;
          DemoLogf("Burst Receiver Hwy[%d] Gamma Test Results (Verbose Form):\n", g_demoEnv.nCurrentHwy);
          for (y = 0; y < g_demoEnv.nBurstHwyTimeslots; y++) {
            DemoLogf("Hwy[%.1d] TS%.3d:  ", g_demoEnv.nCurrentHwy, y);
            DemoLogf("%d", g_hLogicDevs.mBurstReceiver[g_demoEnv.nCurrentHwy].nGammaTestErrors[y]);
            DemoLogf("\n");
          }
        } else {
          DemoLogf("Gamma Test Mode is not enabled\n");
        }
      } else {
        DemoLogf("Burst Receiver Threads are not running. Start them with 'r' or 'R'\n");
      }
    }
    break;

  case 'l':
    {
      OtxT1E1Layer1StatusE eLayer1Status;
      OTX_INT32 i;

      for (i = 0; i < g_demoEnv.coLi; i++) {
        DemoLogf("Line status of Li #%d is ", i);
        result = OtxT1E1GetLayer1Status(g_hPhysDevs.hLi[i], &eLayer1Status); DemoCheckRc(result);
        DemoLogf("%s\n", eLayer1Status == OTX_T1E1_LAYER1_OK ? "OK" : "DOWN");
      }

    }
    break;

  case 'e':
    // by enabling the Burst device, we start the process of transferring data
    // from the hardware to the driver via DMA.  The data is stored in the Burst
    // device's buffer for later processing to connected logical devices.
    DemoLogf("Enable g_hPhysDevs.hBurst\n");
    result = OtxDrvEnable(g_hPhysDevs.hBurst, OTX_TASK_SYNC, 5000L);
    DemoCheckRc(result);
    break;

  case 'd':
    // by disabling the Burst device, we stop the process of transferring data
    // from the hardware to the driver via DMA.
    DemoLogf("Disable g_hPhysDevs.hBurst\n");
    result = OtxDrvDisable(g_hPhysDevs.hBurst, OTX_TASK_SYNC, 5000L);
    DemoCheckRc(result);
    break;

  case 'v':
    g_hLogicDevs.mBurstReceiver[g_demoEnv.nCurrentHwy].bVerbose = !g_hLogicDevs.mBurstReceiver[g_demoEnv.nCurrentHwy].bVerbose;
    DemoLogf("Setting verbose setting to %s for Burst Receiver[%d] Thread\n", (g_hLogicDevs.mBurstReceiver[g_demoEnv.nCurrentHwy].bVerbose)?"ON":"OFF", g_demoEnv.nCurrentHwy);
    break;

  case 'V':
    g_hLogicDevs.mBurstSender[g_demoEnv.nCurrentHwy].bVerbose = !g_hLogicDevs.mBurstSender[g_demoEnv.nCurrentHwy].bVerbose;
    DemoLogf("Setting verbose setting to %s for Burst Sender[%d] Thread\n", (g_hLogicDevs.mBurstSender[g_demoEnv.nCurrentHwy].bVerbose)?"ON":"OFF", g_demoEnv.nCurrentHwy);
    break;

  case 'a':
    {
      OTX_UINT32 nHighwayNo;

      nHighwayNo = g_demoEnv.nCurrentHwy;

      if (g_hLogicDevs.mBurstReceiver[nHighwayNo].bActive) {
        if (g_hLogicDevs.mBurstReceiver[nHighwayNo].nDataBlockCounter == 0) {
          DemoLogf("DemoBurstReceiver[%d]: Error. No burst data was received during the test.\n", nHighwayNo);
          g_nNoRxDataErrors++;
        }
      }

      g_hLogicDevs.mBurstReceiver[nHighwayNo].bActive = OTX_FALSE;
      g_hLogicDevs.mBurstReceiver[nHighwayNo].nRecvCnt = 0;

      if (g_hLogicDevs.mBurstReceiver[nHighwayNo].pRxFile) {
        DemoLogf("DemoBurstReceiver[%d]: Closing file\n", nHighwayNo);
        fclose(g_hLogicDevs.mBurstReceiver[nHighwayNo].pRxFile);
        g_hLogicDevs.mBurstReceiver[nHighwayNo].pRxFile = NULL;
      }

      g_hLogicDevs.mBurstSender[nHighwayNo].bActive   = OTX_FALSE;
      if (g_hLogicDevs.mBurstSender[nHighwayNo].pTxFile) {
        DemoLogf("mBurstSender[%d]: Closing file\n", nHighwayNo);
        fclose(g_hLogicDevs.mBurstSender[nHighwayNo].pTxFile);
        g_hLogicDevs.mBurstSender[nHighwayNo].pTxFile = NULL;
      }

      DemoLogf("Burst Hwy[%d] abort.\n", nHighwayNo);

    }
    break;

  case 'r':
    {
      OTX_UINT32 nHighwayNo;

      nHighwayNo = g_demoEnv.nCurrentHwy;

      if (g_hLogicDevs.mBurstReceiver[nHighwayNo].bActive == OTX_FALSE) {
        DemoLogf("Burst Receiver Hwy[%d]: Receiving data (no file logging).\n", nHighwayNo);
        g_hLogicDevs.mBurstReceiver[nHighwayNo].nDataBlockCounter = 0;
        g_hLogicDevs.mBurstReceiver[nHighwayNo].bActive = OTX_TRUE;
      } else if (g_hLogicDevs.mBurstReceiver[nHighwayNo].pRxFile) {
        DemoLogf("Burst Receiver Hwy[%d]: Closing file, but continuing to receive data.\n", nHighwayNo);
        fclose(g_hLogicDevs.mBurstReceiver[nHighwayNo].pRxFile);
        g_hLogicDevs.mBurstReceiver[nHighwayNo].pRxFile = NULL;
      } else {
        DemoLogf("Burst Receiver Hwy[%d] already running.\n", nHighwayNo);
      }
    }
    break;

  case 'R':
    {
      char sqzBuf[OTX_MAX_PATH];
      char sqzFileName[OTX_MAX_PATH];
      OTX_UINT32 nHighwayNo;
      static OTX_INT32 nTsToRecord = -1;

      nHighwayNo = g_demoEnv.nCurrentHwy;

      if (g_hLogicDevs.mBurstReceiver[nHighwayNo].bActive == OTX_FALSE || g_hLogicDevs.mBurstReceiver[nHighwayNo].pRxFile == NULL) {
        DemoLogf("Burst Receiver Hwy[%d]. Please enter file to receive: ", nHighwayNo);
        OsaConsoleGetStr(sqzBuf);
        if (sqzBuf[0] == '\0') {
          DemoLogf("File name not entered. Operation aborted\n");
          break;
        } 

        strcpy(sqzFileName, sqzBuf);

        if ((g_hLogicDevs.mBurstReceiver[nHighwayNo].pRxFile = fopen(sqzFileName, "wb")) == NULL) {
          DemoLogf("Error: Could not open file %s \n", sqzBuf);
          break;
        }

        DemoLogf("Burst Receiver Hwy[%d]. Please enter single TS[0-%d] to receive to file (-1=All) [%d]: ", nHighwayNo, g_demoEnv.nBurstHwyTimeslots, nTsToRecord);
        OsaConsoleGetStr(sqzBuf);
        if (sqzBuf[0]) {
          nTsToRecord = strtol(sqzBuf, NULL, 0);
        }
        g_hLogicDevs.mBurstReceiver[nHighwayNo].nTsToRecord = nTsToRecord;

        DemoLogf("Burst Receiver Hwy[%d]. Receiving data (TS=%d) to file \"%s\". Abort using command 'a'.\n", nHighwayNo, nTsToRecord, sqzFileName);
        g_hLogicDevs.mBurstReceiver[nHighwayNo].nDataBlockCounter = 0;
        g_hLogicDevs.mBurstReceiver[nHighwayNo].bActive = OTX_TRUE;
      } else {
        DemoLogf("Burst Receiver Hwy[%d] Thread already running.\n", nHighwayNo);
      }
    }
    break;

  case 's':
    {
      OTX_UINT32 nHighwayNo, i;

      nHighwayNo = g_demoEnv.nCurrentHwy;

      if (g_hLogicDevs.mBurstSender[nHighwayNo].bActive == OTX_FALSE) {
        DemoLogf("Burst Sender Hwy[%d]. Sending a gamma pattern continuously\n", nHighwayNo);

        g_hLogicDevs.mBurstSender[nHighwayNo].pTxFile = NULL;

        // The starting point for the gamma pattern will be value=timeslot
        g_hLogicDevs.mBurstSender[nHighwayNo].cbTxBuf = 0;
        for (i = 0; i < (OTX_UINT32)g_demoEnv.nBurstHwyTimeslots; i++) {
          //g_hLogicDevs.mBurstSender[nHighwayNo].sqGammaData[i] = (unsigned char) i;
          g_hLogicDevs.mBurstSender[nHighwayNo].sqGammaData[i] = (unsigned char) nHighwayNo;
        }

        g_hLogicDevs.mBurstSender[nHighwayNo].bActive = OTX_TRUE;

        g_hLogicDevs.mBurstSender[nHighwayNo].cbTxBuf = DEMO_BURST_BUF_SIZE;
      } else {
        DemoLogf("Burst Sender Hwy[%d] already running.\n", nHighwayNo);
      }
    }

    break;

  case 'S':
    {
      char sqzBuf[OTX_MAX_PATH];
      char sqzFileName[OTX_MAX_PATH];
      OTX_UINT32 nHighwayNo;
      static OTX_INT32 nTsToSend = -1;

      nHighwayNo = g_demoEnv.nCurrentHwy;
      if (g_hLogicDevs.mBurstSender[nHighwayNo].bActive == OTX_FALSE) {
        DemoLogf("Burst Sender Hyw[%d]. Please enter file to send: ", nHighwayNo);

        OsaConsoleGetStr(sqzFileName);
        if (sqzFileName[0] == '\0') {
          DemoLogf("File name not entered. Operation aborted\n");
          break;
        }

        if ((g_hLogicDevs.mBurstSender[nHighwayNo].pTxFile = fopen(sqzFileName, "rb")) == NULL) {
          DemoLogf("Error: Could not open file %s \n", sqzFileName);
          break;
        }

        DemoLogf("Burst Sender Hwy[%d]. Please enter single TS[0-%d] to send from file (-1=All) [%d]: ", nHighwayNo, g_demoEnv.nBurstHwyTimeslots, nTsToSend);
        OsaConsoleGetStr(sqzBuf);
        if (sqzBuf[0]) {
          nTsToSend = strtol(sqzBuf, NULL, 0);
        }
        g_hLogicDevs.mBurstSender[nHighwayNo].nTsToSend = nTsToSend;

        DemoLogf("Burst Sender Hwy[%d]. Sending data (TS=%d) from file \"%s\". Abort using command 'a'.\n", nHighwayNo, nTsToSend, sqzFileName);
        g_hLogicDevs.mBurstSender[nHighwayNo].bActive = OTX_TRUE;
        g_hLogicDevs.mBurstSender[nHighwayNo].bVerbose = OTX_TRUE;
      } else {
        DemoLogf("Burst Sender Hwy[%d] already running.\n", nHighwayNo);
      }
    }
    break;

  case 'L':
    g_hLogicDevs.mBurstSender[g_demoEnv.nCurrentHwy].bFileLoop = !g_hLogicDevs.mBurstSender[g_demoEnv.nCurrentHwy].bFileLoop;
    DemoLogf("Setting loop file sends %s for Burst Highway[%d] Thread\n", (g_hLogicDevs.mBurstSender[g_demoEnv.nCurrentHwy].bFileLoop) ? "ON" : "OFF", g_demoEnv.nCurrentHwy);
    break;


  case 'f':
    {
      OTX_UINT32 nBytesReturned;
      Demo2IoOutGetFullFrameS dataS;

      if (g_hUserApp[g_demoEnv.nCurrentHwy] != OTX_INVALID_HANDLE_VALUE) {
        
        DemoLogf("Full Frame sample: DSP[%d]\n", g_demoEnv.nCurrentHwy);
        
        result = OtxDspIoControl(g_hUserApp[g_demoEnv.nCurrentHwy], DEMO2_IOCTL_GET_FULL_FRAME, NULL, 0, &dataS, sizeof(dataS), &nBytesReturned, OTX_TASK_SYNC, OTX_TIME_INFINITY); DemoCheckRc(result);
        
        if (OTX_SUCCESS(result)) {
          if (nBytesReturned != sizeof(Demo2IoOutGetFullFrameS)) {
            DemoLogf("Unexpected data length for DEMO2_IOCTL_GET_FULL_FRAME (nBytesReturned=%d)\n", nBytesReturned);
          } else if (dataS.nResult != OTXDSP_SDK_S_OK) {
            DemoLogf("Demo2 Error Code: %d\n", dataS.nResult);
          } else {
            OTX_INT32 i;
            for (i = 0; i < (OTX_INT32)g_nDspHighwaySpeed/64; i++) {
              if ((i != 0) && ((i%8) == 0)) DemoLogf("\n");
              DemoLogf("[%.3d]:%.2X ", i, (OTX_UINT8)dataS.nFrame[i]);
            }
            DemoLogf("\n");
            
          }
        }
      }

    }
    break;


  case 'M':
    {
      DemoLogf("Dsp[%d] Mem Test\n", g_demoEnv.nCurrentHwy);
      OtxDrvDbgDump(g_hDsp[g_demoEnv.nCurrentHwy], 4);
     }
    break;

  case 'O':
    {
      char szBuf[64];
      OTX_INT32 nCmd;
      
      DemoLogf("Driver Debug Command to the board device. Which Cmd?\n");
      DemoLogf("0x2000: Codec Normal operation\n");
      DemoLogf("0x2001: Digital loopback of all Codec channels\n");
      DemoLogf("0x2002: TSA loopback of all Codec channels\n");
      DemoLogf("0x2003: 1 kHz receive tone on\n");
      OsaConsoleGetStr(szBuf);
      nCmd = (OTX_INT32) strtol(szBuf, NULL, 0);
      
      DemoLogf("Issuing Driver Debug Command 0x%X to board device\n", nCmd);
      OtxDrvDbgDump(g_hPhysDevs.hThorController, nCmd);
    }
    break;

  case '{':
    {
      char szBuf[64];
      OTX_INT32 nCmd;
      static OTX_INT32 nCh = 0;
      static OTX_INT32 nReg = 0x70;
      static OTX_INT32 nData;
      
      DemoLogf("Read Register from Codec.\n");
      DemoLogf("Enter Channel[%d]: ", nCh);
      OsaConsoleGetStr(szBuf);
      if (szBuf[0]) {
        nCh = (OTX_INT32) strtol(szBuf, NULL, 0);
      }
      DemoLogf("Enter Register[0x%.2X]: ", nReg);
      OsaConsoleGetStr(szBuf);
      if (szBuf[0]) {
        nReg = (OTX_INT32) strtol(szBuf, NULL, 0);
      }
      
      nCmd = 0x3010 | nCh | (nReg << 24);
      DemoLogf("Reading Codec Ch=%d, Register=0x%.2X\n", nCh, nReg);
      OtxDrvDbgDump(g_hPhysDevs.hThorController, nCmd);
    }
    break;

  case '}':
    {
      char szBuf[64];
      OTX_INT32 nCmd;
      static OTX_INT32 nCh = 0;
      static OTX_INT32 nReg = 0x70;
      static OTX_INT32 nData = 0x00;
      static OTX_INT32 nData2 = 0x00;
      
      DemoLogf("Write Register to Codec.\n");
      DemoLogf("Enter Channel[%d]: ", nCh);
      OsaConsoleGetStr(szBuf);
      if (szBuf[0]) {
        nCh = (OTX_INT32) strtol(szBuf, NULL, 0);
      }
      DemoLogf("Enter Register[0x%.2X]: ", nReg);
      OsaConsoleGetStr(szBuf);
      if (szBuf[0]) {
        nReg = (OTX_INT32) strtol(szBuf, NULL, 0);
      }
      if ((nReg >=0x80) && (nReg <=0x82)) {
        // Dual data byte command
        DemoLogf("Enter Data1[0x%.2X]: ", nData);
        OsaConsoleGetStr(szBuf);
        if (szBuf[0]) {
          nData = (OTX_INT32) strtol(szBuf, NULL, 0);
        }
        DemoLogf("Enter Data2[0x%.2X]: ", nData2);
        OsaConsoleGetStr(szBuf);
        if (szBuf[0]) {
          nData2 = (OTX_INT32) strtol(szBuf, NULL, 0);
        }
        nCmd = 0x3020 | nCh | (nData << 24) | (nData2 << 16);
        if (nReg == 0x82) nCmd |= 0x10;
        DemoLogf("Writing Codec Ch=%d, Register=0x%.2X, Data1=0x%.2X, Data2=0x%.2X\n", nCh, nReg, nData, nData2);
      } else {
        // Single data byte command
        DemoLogf("Enter Data[0x%.2X]: ", nData);
        OsaConsoleGetStr(szBuf);
        if (szBuf[0]) {
          nData = (OTX_INT32) strtol(szBuf, NULL, 0);
        }

        nCmd = 0x3000 | nCh | (nReg << 24) | (nData << 16);
        DemoLogf("Writing Codec Ch=%d, Register=0x%.2X, Data=0x%.2X\n", nCh, nReg, nData);
      }
      
      OtxDrvDbgDump(g_hPhysDevs.hThorController, nCmd);
    }
    break;

  case 'F':
    {
      OTX_UINT32 nBytesReturned;
      DemoLogf("Get Debug Data for User App #%d (non-blocking call)...", g_demoEnv.nCurrentHwy);

      result = OtxDspIoControl(g_hUserApp[g_demoEnv.nCurrentHwy], DEMO2_IOCTL_GET_DEBUG_DATA, NULL, 0, NULL, 0, &nBytesReturned, 0x1234, 5000); DemoCheckRc(result);

      DemoLogf("Done (nBytesReturned=%d).\n", nBytesReturned);
    }
    break;

  case 'c':
    {
      char szBuf[64];
      static OTX_INT32 nTransferType;
      szBuf[0] = 0;
      
      DemoLogf("H.100 Transmit or Recieve\n");
      DemoLogf("Enter R or T : ");
      OsaConsoleGetStr(szBuf);
      if(szBuf[0]) {
        if((szBuf[0] == 'R') || (szBuf[0] == 'r')) {
          // Burst device will receive data from H.100
          nTransferType = 1; // Recieve data from H.100
          DemoConnectH100(nTransferType);
        }
        else
        if((szBuf[0] == 'T') || (szBuf[0] == 't')) {
          // Burst device will transmit data to H.100
          nTransferType = 0; // Send data to H.100
          DemoConnectH100(nTransferType);
        }
        else {
          DemoLogf("Abort entry - %c\n", szBuf[0]);
        }
      }
      else {
        DemoLogf("Abort no entry\n");
      }
    }
    break;

  case 'C':
    {
      OTX_UINT8 highWay;
      OTX_UINT8 channel;
      OTX_UINT8 byteVal;
      char wValStr[12];
      
      DemoLogf("Set Constant Byte on Channel\n");
      
      DemoLogf("Highway: ");
      OsaConsoleGetStr(wValStr);
      highWay = (OTX_UINT8) strtoul(wValStr, NULL, 0);
      
      DemoLogf("Channel: ");
      OsaConsoleGetStr(wValStr);
      channel = (OTX_UINT8) strtoul(wValStr, NULL, 0);
      
      DemoLogf("Constant Byte: ");
      OsaConsoleGetStr(wValStr);
      byteVal = (OTX_UINT8) strtoul(wValStr, NULL, 0);
      
      result = OtxBrdTssConstByte(g_hPhysDevs.hTss, highWay, channel, byteVal); 
      DemoCheckRc(result);

      DemoLogf("Done.\n");
    }
    break;

  case '0':
  case '1':
    if (commandChar - '0' < g_demoEnv.coBurstHighways) {
      // Select current Li
      g_demoEnv.nCurrentHwy = commandChar - '0';
      DemoLogf("HWY#%d Selected ....\n", g_demoEnv.nCurrentHwy);
    }
    break;

  case 'q':
  case 'Q':
    return 1;
    break;

  case '~':
    {
      char strBuf[12];
      OTX_INT32 nSleepTime;

      DemoPrintf("Enter milliseconds to sleep :");
      OsaConsoleGetStr(strBuf);

      if (strBuf[0]) {
        nSleepTime = strtol(strBuf, NULL, 0);
        DemoPrintf("Sleeping %d milliseconds\n", nSleepTime);
        OsaThreadSleep(nSleepTime);
      }
    }
    break;

  default:
    DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
    break;
  }

  return 0;
}




/*******************************************************************************

  @func

  Parse the program's command line arguments and calls upon
  requested actions.

  There are some optional switches that can be used following
  the driver name on the command line. Switches are are
  separated by spaces, and can appear in arbitrary order.
  Switches consist of a hyphen and a letter:

  @comm

*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP)
{
  short  i;

  // Set reasonable defaults before reading from command line
  envP->pszInputStr       = NULL;
  envP->bExitOnError      = OTX_FALSE;
  envP->boardNo           = DEMO_DEFAULT_BOARD_NO;
  envP->nBoardType        = OTX_DEVICE_THOR_8_PCI_PLUS;
  envP->coBurstHighways        = 2;
  envP->nBurstHwyTimeslots     = 128;
  envP->bDisableBurstTx   = OTX_FALSE;
  envP->bDisableBurstRx   = OTX_FALSE;

  envP->bUseE1Loop       = OTX_FALSE;
  envP->bUseExternalLoop = OTX_FALSE;
  envP->bUseDspLoop      = OTX_FALSE;
  envP->bUseH100Loop     = OTX_FALSE;
  envP->nHbusRate        = 2048;
  envP->nH100HwyRate     = 2048;
  envP->eLaw             = OTX_DATA_FORMAT_COMPAND8_A_LAW;
  envP->nDspNo           = 1;

  envP->eClkSource        = OTX_CLOCK_SOURCE_INTERNAL;

  strcpy(envP->logFileName, envP->cwd);
  strcat(envP->logFileName, DEMO_DEFAULT_LOG_FILE);
  for (i = 1; i < argc; i++) {        // Scan command line arguments
    if (*(argv[i]+1) == 'Z') {   // Command input string -- case sensitive
      envP->pszInputStr = argv[i]+2;
    } else {
      OsaStdStrUpr(argv[i]);
      if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
        if (*(argv[i]+1) == 'L') {   // Log file
          strcpy(envP->logFileName, argv[i] + 2);
        } else if (*(argv[i]+1) == 'Q') {   // Exit on error
          envP->bExitOnError = OTX_TRUE;
        } else if (*(argv[i]+1) == 'B') {   // Board number
          envP->boardNo = strtoul(argv[i]+2, NULL, 0);
        } else if (*(argv[i]+1) == 'X') {   // Use External loop with loop cables
          envP->bUseExternalLoop = OTX_TRUE;
        } else if (*(argv[i]+1) == 'H') {   // Use H100 loop with loop cables
          envP->bUseH100Loop = OTX_TRUE;
        } else if (*(argv[i]+1) == 'P') {   // Use loop via DSP
          envP->bUseDspLoop = OTX_TRUE;
        } else if (*(argv[i]+1) == 'C') {   // Use External Codec loop with Codec Loop plug; Companding Law
          envP->bUseCodecLoop    = OTX_TRUE;
          if (*(argv[i]+2) == 'A') {   // A-law
            envP->eLaw = OTX_DATA_FORMAT_COMPAND8_A_LAW;
          } else if (*(argv[i]+2) == 'U') {   // U-law
            envP->eLaw = OTX_DATA_FORMAT_COMPAND8_U_LAW;
          } else {
            DemoLogf("Error: Invalid companding law command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
					  exit(1);
          }
        } else if (*(argv[i]+1) == 'T') {   // Board Type
          if ( *(argv[i]+2) == '4') {
            envP->nBoardType = OTX_DEVICE_THOR_4_PCI_PLUS;
            envP->coBurstHighways = 1;
            envP->nBurstHwyTimeslots = 128;
            envP->nH100HwyRate = 2048;
          } else if ( *(argv[i]+2) == 'P') {
            envP->nBoardType = OTX_DEVICE_THOR_8_PCI_PLUS;
            envP->coBurstHighways = 2;
            envP->nBurstHwyTimeslots = 128;
            envP->nH100HwyRate = 2048;
          } else if ( *(argv[i]+2) == 'T') {
            envP->nBoardType = OTX_DEVICE_THOR_2_PCI_PLUS;
            envP->coBurstHighways = 2;
            envP->nBurstHwyTimeslots = 32;
            envP->nH100HwyRate = 2048;
          } else if ( *(argv[i]+2) == 'E') {
            envP->nBoardType = OTX_DEVICE_THOR_2_PCI_EXPRESS;
            envP->coBurstHighways = 2;
            envP->nBurstHwyTimeslots = 32;
            envP->nH100HwyRate = 2048;
          } else if ( *(argv[i]+2) == '1') {
            envP->nBoardType = OTX_DEVICE_THOR_1_PCI_PLUS;
            envP->coBurstHighways = 1;
            envP->nBurstHwyTimeslots = 32;
            envP->nH100HwyRate = 2048;
          } else if ( *(argv[i]+2) == 'C') {
            envP->nBoardType = OTX_DEVICE_THOR_2_EXPRESSCARD;
            envP->coBurstHighways = 1;
            envP->nBurstHwyTimeslots = 128;
            envP->nH100HwyRate = 0; // not supported on this card
          } else if ( *(argv[i]+2) == 'R') {
            envP->nBoardType = OTX_DEVICE_THOR_4_EXPRESSCARD;
            envP->coBurstHighways = 1;
            envP->nBurstHwyTimeslots = 128;
            envP->nH100HwyRate = 0; // not supported on this card
          } else if ( *(argv[i]+2) == 'H') {
            envP->nBoardType = OTX_DEVICE_HUGIN_2_PC104;
            envP->nDspNo          = 2;
            envP->coBurstHighways = 2;
            envP->nBurstHwyTimeslots = 32;
            envP->bUseDspLoop = OTX_TRUE;
            envP->nH100HwyRate = 0; // not supported on this card
          }
          else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
        } else if (*(argv[i]+1) == 'O') {   // Clock Source
          if (*(argv[i]+2) == 'I') {   // Internal
            envP->eClkSource = OTX_CLOCK_SOURCE_INTERNAL;
          } else if (*(argv[i]+2) == 'H') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_A_CLOCKS_ETCF;
          } else {
            DemoLogf("Error: Invalid Clock Source command line switch %s\n", argv[i]);
            DemoHelpPanel();
            exit(1);
          }
          
        } else if (*(argv[i]+1) == 'D') {   // Disable one direction of burst
          if (*(argv[i]+2) == 'T') {   // TX
            envP->bDisableBurstTx = OTX_TRUE;
          } else if (*(argv[i]+2) == 'R') {   // T1
            envP->bDisableBurstRx = OTX_TRUE;
          } else {
            DemoLogf("Error: Invalid Burst direction-disabling command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
        } else if (*(argv[i]+1) == 'R') {   // Hbus Rate
          envP->nHbusRate = strtol(argv[i]+2, NULL, 0);
          switch (envP->nHbusRate) {
          case 2048:
          case 4096:
          case 8192:
            break; // Acceptable value
          
          default:
            DemoLogf("Error: Invalid Hbus Rate %d. Valid rates are 2048, 4096 and 8192 kbps.\n", envP->nHbusRate);
            DemoHelpPanel();
            OsaConsoleExit();
					  exit(1);
          }
        } else {
          DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          OsaConsoleExit();
          exit(1);
        }
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
        exit(1);
      }
    }
  }
}


/*******************************************************************************

  @func

  Displays a help for the command line.

  @comm

*******************************************************************************/
static void DemoHelpPanel()
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /L<fileName> : Log file name\n");
  DemoLogf(" /B<boardNo>  : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /T[T|E|4|P|1|C|R|H] : Board Type used for the demo, where\n");
  DemoLogf("                  T = Use Thor-2-PCI-Plus Board\n");
  DemoLogf("                  E = Use Thor-2-PCI-Express Board\n");
  DemoLogf("                  4 = Use Thor-4-PCI-Plus Board\n");
  DemoLogf("                  P = Use Thor-8-PCI-Plus Board\n");
  DemoLogf("                  1 = Use Thor-1-PCI-Plus Board\n");
  DemoLogf("                  C = Use Thor-2-ExpressCard Board\n");
  DemoLogf("                  R = Use Thor-2-ExpressCard Board\n");
  DemoLogf("                  H = Use Hugin-2-PC104 Board\n");
  DemoLogf(" /D[T|R]      : Disable one direction of the Burst device\n");
  DemoLogf("                  T = Disable Burst TX\n");
  DemoLogf("                  R = Disable Burst RX\n");
  DemoLogf(" /CA          : Use loop via Codec in a-law mode (needs Codec loop plug)\n");
  DemoLogf(" /CU          : Use loop via Codec in u-law mode (needs Codec loop plug)\n");
  DemoLogf(" /P           : Use loop via DSP \n");
  DemoLogf(" /X           : Use external loop cables (default is internal looping)\n");
  DemoLogf(" /H           : Use H100 loop cables (default is internal looping)\n");
  DemoLogf(" /R[2048|4096|8192] : Use H100 Rate (default 2048)\n");
  DemoLogf(" /Q           : Exit without prompting on errors\n");
  DemoLogf(" /O[I|H]      : Clock Source, where\n");
  DemoLogf("                  I Internal clock source (default) or\n");
  DemoLogf("                  H H100 interface\n");
}


/*******************************************************************************

  @func

  Displays a message on the screen and prints it to the logfile.

  @comm

*******************************************************************************/
static void DemoLogf(char *formatStr, ...)
{
  va_list argList;

  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);

  if (g_pDemoLogFile) {
    vfprintf(g_pDemoLogFile,   formatStr, argList);

    fflush(g_pDemoLogFile);
  }
  va_end(argList);

}


/*******************************************************************************
*******************************************************************************/
void DemoAtExit(void)
{
  OTX_INT32 i;

  for (i = 0; i < g_demoEnv.coBurstHighways; i++) {
    if (g_hLogicDevs.mBurstSender[i].pTxFile) {
      fclose(g_hLogicDevs.mBurstSender[i].pTxFile);
      g_hLogicDevs.mBurstSender[i].pTxFile = NULL;
    }
    if (g_hLogicDevs.mBurstReceiver[i].pRxFile) {
      fclose(g_hLogicDevs.mBurstReceiver[i].pRxFile);
      g_hLogicDevs.mBurstReceiver[i].pRxFile = NULL;
    }
  }

  if (g_pDemoLogFile) {
    fclose(g_pDemoLogFile);
  }


}

int main(int argc, char *argv[])
/*******************************************************************************
*******************************************************************************/
{
  OTX_RESULT result;  // OTX Standard Return Code
  char commandChar;
  int i;

  OsaConsoleInit();

  for(i=0; i<DEMO_MAX_DSP; i++) {
    g_hDsp[i]          = OTX_INVALID_HANDLE_VALUE;
    g_hUserApp[i]      = OTX_INVALID_HANDLE_VALUE;
  }
  for(i=0; i<8; i++) {
    g_hPhysDevs.hLi[i] = OTX_INVALID_HANDLE_VALUE;
  }

  printf("T8pBurstBasic, %s, Odin TeleSystems Inc.\n", __DATE__);
  printf("Example program for Burst device.\n");

  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if (g_demoEnv.pszInputStr) {
    OsaConsoleInput(g_demoEnv.pszInputStr);
  }

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
  }
  atexit(DemoAtExit);

  DemoLogf("Logg file is %s\n", g_demoEnv.logFileName);

  // Query Library revision
  {
    char szVersionInfo[255];
    OTX_BOOL bIsStaticLinkedLibrary;
    char szLibFileName[64];
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_REV_STR, szVersionInfo, 255); DemoCheckRc(result);
    result = OtxDrvGetAttributeValueBool(OTX_LIB_HANDLE, OTX_ATTR_LIB_TYPE_STATIC, &bIsStaticLinkedLibrary); DemoCheckRc(result);
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_FILE_NAME, szLibFileName, 64); DemoCheckRc(result);

    DemoLogf("Library version %s (filename %s, %s linked)\n", szVersionInfo, szLibFileName, bIsStaticLinkedLibrary ? "static":"dynamic");
  }


  // Connect to the Driver
  DemoLogf("Connecting lib...\n");
  result = OtxDrvConnectLib();
  switch (OTX_RESULT_CODE(result)) {
  case OTX_RES_HANDLE:
    DemoLogf("Error: Could not connect to the OTX library.\nPossible cause: OTX Driver not loaded\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  case OTX_RES_NOT_SUPPORTED:
    DemoLogf("Error: Mismatch between OTX Driver revision and OTX Library revision.\nRevisions must match.\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  default:
    DemoCheckRc(result);
    break;
  }

  // Create Event Queues for Driver --> Application Communications
  DemoLogf("DemoCreateEventQueue\n");
  DemoCreateEventQueue();

  // Connect (get handles) to the physical devices on the Thor-2-PCMCIA board
  DemoLogf("DemoCreatePhysicalDevices\n");
  DemoCreatePhysicalDevices();

  // Enable physical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("DemoEnablePhysicalDevices\n");
  DemoEnablePhysicalDevices();
#if 0
  // Set clock source for the board
  result = OtxBrdSetClocks(g_hPhysDevs.hThorController, OTX_CLOCK_MODE_H100_ETCF, OTX_CLOCK_SOURCE_INTERNAL); DemoCheckRc(result);
#else
  // Set clock source for the board
  {
    char szClockSource[64];

    switch (g_demoEnv.eClkSource) {
    case OTX_CLOCK_SOURCE_INTERNAL:
      strcpy(szClockSource, "Internal");
      break;
    case OTX_CLOCK_SOURCE_A_CLOCKS_ETCF:
      sprintf(szClockSource, "H.100 A-Clocks (ETCF mode)");
      break;
    case OTX_CLOCK_SOURCE_LOCAL_0:
    case OTX_CLOCK_SOURCE_LOCAL_1:
    case OTX_CLOCK_SOURCE_LOCAL_2:
    case OTX_CLOCK_SOURCE_LOCAL_3:
    case OTX_CLOCK_SOURCE_LOCAL_4:
    case OTX_CLOCK_SOURCE_LOCAL_5:
    case OTX_CLOCK_SOURCE_LOCAL_6:
    case OTX_CLOCK_SOURCE_LOCAL_7:
      sprintf(szClockSource, "Line Interface Clock Source not supported");
      break;    default:
      strcpy(szClockSource, "Unknown");
      break;
    }

    DemoLogf("Clock Source: %s\n", szClockSource);

    result = OtxBrdSetClocks(g_hPhysDevs.hThorController, OTX_CLOCK_MODE_H100_ETCF, g_demoEnv.eClkSource); DemoCheckRc(result);
  }
#endif

  // Get a handle to the physical TSS device.
  result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController, OTX_DEVICE_TSS, 0, 0, g_hEventQueue, &g_hPhysDevs.hTss); DemoCheckRc(result);

  
  // Connect to the Physical CT Buses
  if(g_demoEnv.bUseH100Loop) {
    DemoLogf("Open OTX_DEVICE_CT_BUS \n");
    result = OtxDrvOpenPhysicalDevice(0, 
                                      OTX_DEVICE_CT_BUS, 
                                      0, 
                                      0, 
                                      g_hEventQueue, 
                                      &g_hPhysDevs.hCtBus); 
    DemoCheckRc(result);

    // NOTE: need to increase the following attributes when necessary
    // Limit the number of highways that OTX can use on this CT Bus 
    // (only to demonstrate this functionality)
    result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hCtBus, OTX_ATTR_CTBUS_OTX_HWYS, 4); 
    DemoCheckRc(result);
    // Limit the number of timeslots that OTX can use on this CT Bus 
    // (only to demonstrate this functionality)
    result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hCtBus, OTX_ATTR_CTBUS_OTX_TIMESLOTS_PER_HWY, 32); 
    DemoCheckRc(result);

    result = OtxDrvConnectBoard(g_hPhysDevs.hCtBus, g_hPhysDevs.hThorController); 
    DemoCheckRc(result);

    // If we do not close the device the other cards will not be able to 
    // connect to the bus
    OtxDrvClose(g_hPhysDevs.hCtBus);

    // Set the H.100 Rate (default is 2048 kbps)    
    DemoLogf("Setting rate to %d for all H.100 highways.\n", g_demoEnv.nHbusRate);
    result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hTss, OTX_ATTR_TSS_CTBUS_RATE_STREAM_GROUP_A, g_demoEnv.nHbusRate); DemoCheckRc(result);
    result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hTss, OTX_ATTR_TSS_CTBUS_RATE_STREAM_GROUP_B, g_demoEnv.nHbusRate); DemoCheckRc(result);
    result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hTss, OTX_ATTR_TSS_CTBUS_RATE_STREAM_GROUP_C, g_demoEnv.nHbusRate); DemoCheckRc(result);
    result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hTss, OTX_ATTR_TSS_CTBUS_RATE_STREAM_GROUP_D, g_demoEnv.nHbusRate); DemoCheckRc(result);
    result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hTss, OTX_ATTR_TSS_CTBUS_RATE_STREAM_GROUP_E, g_demoEnv.nHbusRate); DemoCheckRc(result);
    result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hTss, OTX_ATTR_TSS_CTBUS_RATE_STREAM_GROUP_F, g_demoEnv.nHbusRate); DemoCheckRc(result);
    result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hTss, OTX_ATTR_TSS_CTBUS_RATE_STREAM_GROUP_G, g_demoEnv.nHbusRate); DemoCheckRc(result);
    result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hTss, OTX_ATTR_TSS_CTBUS_RATE_STREAM_GROUP_H, g_demoEnv.nHbusRate); DemoCheckRc(result);

    DemoLogf("Enable OTX_DEVICE_TSS\n");
    result = OtxDrvEnable(g_hPhysDevs.hTss, OTX_TASK_SYNC, 5000L); 
    DemoCheckRc(result);
  }
 
  // Create and get handles to the Pipe devices needed by this application
  DemoLogf("DemoConnectDevices\n");
  DemoConnectDevices();

  // Read Serial Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszSerialNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController, OTX_ATTR_BRD_SERIAL_NUMBER, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszSerialNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController, OTX_ATTR_BRD_SERIAL_NUMBER, pszSerialNumber, cbAttribValue); DemoCheckRc(result);
    DemoLogf("Board Serial Number: %s\n", pszSerialNumber);
    free(pszSerialNumber);
  }

  // Read Build Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszBrdBuildNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController, OTX_ATTR_BRD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszBrdBuildNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController, OTX_ATTR_BRD_BUILD, pszBrdBuildNumber, cbAttribValue); DemoCheckRc(result);
    DemoLogf("Board Build Number: %s\n", pszBrdBuildNumber);
    free(pszBrdBuildNumber);
  }

  {
    // Read CPLD Build Numbers from the board
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszCpldBuildNumbers;
    
    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController, OTX_ATTR_BRD_CPLD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszCpldBuildNumbers = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController, OTX_ATTR_BRD_CPLD_BUILD, pszCpldBuildNumbers, cbAttribValue); DemoCheckRc(result);
    DemoLogf("CPLD Build:\n%s", pszCpldBuildNumbers);
    free(pszCpldBuildNumbers);
  }

  // Main Loop. Loop here and wait for Events from the driver or commands from the user ....
  OsaThreadStart(DemoIdleFunction);

  OsaThreadSleep(300); // Only so the following message will show on the screen

  DemoLogf("\n");
  DemoLogf("Press 'h' for help\n");

  while (!g_exitFlag) {
    commandChar = OsaConsoleGetCh();
    g_exitFlag = DemoProcessUserCommand(commandChar);
  }

  // Abort threads
  DemoLogf("\n");
  DemoLogf("Exiting...\n");
  {
    OTX_INT32 i;

    g_bStatsThreadActive = OTX_FALSE;
    g_bReadBurstStatsThreadActive = OTX_FALSE;
    for (i = 0; i < g_demoEnv.coBurstHighways; i++) {
      g_hLogicDevs.mBurstReceiver[i].bActive = OTX_FALSE;
      g_hLogicDevs.mBurstSender[i].bActive = OTX_FALSE;
    }

    //OsaThreadSleep(1500);   // give threads a chance to exit
  }

  {
    int m;
    int n;
    int nDataMismatchErrors = 0;
    for (m = 0; m < g_demoEnv.coBurstHighways; m++) {
      for (n = 0; n < g_demoEnv.nBurstHwyTimeslots; n++) {
       if (g_demoEnv.bUseExternalLoop) {
         switch (g_demoEnv.nBoardType) {
         case OTX_DEVICE_THOR_1_PCI_PLUS:
         case OTX_DEVICE_THOR_2_PCI_PLUS:
         case OTX_DEVICE_THOR_2_PCI_EXPRESS:
         case OTX_DEVICE_HUGIN_2_PC104:

           if (n > 0) {
             // The first TS is TS0 from the Li. Since we don't run in unframed mode we need to ignore these errors
             nDataMismatchErrors += g_hLogicDevs.mBurstReceiver[g_demoEnv.nCurrentHwy].nGammaTestErrors[n];
           }
           break;
         case OTX_DEVICE_THOR_4_PCI_PLUS:
         case OTX_DEVICE_THOR_8_PCI_PLUS:
           if (n > 3) {
             // The first four TS are TS0 from the Lis. Since we don't run in unframed mode we need to ignore these errors
             nDataMismatchErrors += g_hLogicDevs.mBurstReceiver[g_demoEnv.nCurrentHwy].nGammaTestErrors[n];
           }
           break;
         case OTX_DEVICE_THOR_2_EXPRESSCARD:
         case OTX_DEVICE_THOR_4_EXPRESSCARD:
           if (n > 3) {
             // The first four TS are TS0 from the Lis. Since we don't run in unframed mode we need to ignore these errors
             nDataMismatchErrors += g_hLogicDevs.mBurstReceiver[g_demoEnv.nCurrentHwy].nGammaTestErrors[n];
           }
           break;
         default:
           DemoLogf("Error: Unexpected board Board Type (%d)\n", g_demoEnv.nBoardType);
           break;
         }
       } else {
         // Make all errors count
         nDataMismatchErrors += g_hLogicDevs.mBurstReceiver[g_demoEnv.nCurrentHwy].nGammaTestErrors[n];
       }

      }
    }

    if (g_nNoRxDataErrors) {
      DemoLogf("Error: At least one highway did not receice any OTX_BURST_RECEIVER_EC_DATA_AVAILABLE events.\n");
    }
    if (nDataMismatchErrors) {
      DemoLogf("Error: A total of %d Data Mismatch Errors occurred\n", nDataMismatchErrors);
    }
    if (g_nUnderRunErrors) {
      DemoLogf("Error: A total of %d Sender Underrun Errors occurred\n", g_nUnderRunErrors);
    }
    if (g_nOverRunErrors) {
      DemoLogf("Error: A total of %d Receiver Overrun Errors occurred\n", g_nUnderRunErrors);
    }
    
    if (g_nUnderRunErrors || g_nOverRunErrors) {
      g_nExitCode = 100;
    } else if (g_nNoRxDataErrors) {
      g_nExitCode = 1;
    } else if (nDataMismatchErrors) {
      g_nExitCode = 2;
    } else {
      g_nExitCode = 0; // Success
    }
  }

  // Clean up
  OsaConsoleExit();
  OtxDrvReleaseLib();

  printf("g_nExitCode = %d\n", g_nExitCode);

  return g_nExitCode;
}
